from __future__ import annotations

import io
import json
from pathlib import Path
import subprocess
import sys
import threading

import pytest

from gotta.actors import ACTOR_SPEAKER_ENV
import gotta.content.activity as content_activity
import gotta.content.context as content_context
import gotta.content.env as content_env
import gotta.content.scope as content_scope
from gotta import friction
from gotta.plugins import actor
from gotta.plugins import oops
from gotta.plugins.session import main as session_plugin
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope


@pytest.fixture(autouse=True)
def local_session_registry(tmp_path: Path, monkeypatch) -> None:
    registry = tmp_path / "sessions"
    monkeypatch.setattr(content_scope, "DEFAULT_SESSION_ROOT", registry)
    monkeypatch.delenv(ACTOR_SPEAKER_ENV, raising=False)
    monkeypatch.delenv(content_env.ACTOR_ID_ENV, raising=False)
    monkeypatch.delenv(content_env.SESSION_ACTOR_ENV, raising=False)


def initialize_session(root: Path) -> None:
    assert session_plugin.main(["init", "--session", str(root)]) == 0


def _run_oops_subprocess(
    argv: list[str],
    *,
    stdin_text: str | None = None,
    closed_stdin: bool = False,
) -> subprocess.CompletedProcess[str]:
    repo_root = Path(__file__).resolve().parents[1]
    command = [
        sys.executable,
        "-c",
        (
            "import sys; "
            "from gotta.plugins import oops; "
            "raise SystemExit(oops.main(sys.argv[1:]))"
        ),
        *argv,
    ]
    kwargs: dict[str, object] = {
        "cwd": repo_root,
        "capture_output": True,
        "text": True,
    }
    if closed_stdin:
        kwargs["stdin"] = subprocess.DEVNULL
    else:
        kwargs["input"] = stdin_text
    return subprocess.run(command, **kwargs)


def test_oops_help_all_is_top_level(capsys) -> None:
    assert oops.main(["--help-all"]) == 0
    output = capsys.readouterr().out
    assert "usage: gotta oops" in output
    assert "Show or record durable session speed bumps" in output
    assert (
        "Bare multiword prose, real piped stdin, `--stdin`, and `--from-file` imply"
        in output
    )


def test_oops_is_session_rooted_and_canonical(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert (
        oops.main(
            [
                "append",
                "Need clearer next step",
                "--session",
                str(root),
                "--surface",
                "read",
                "--kind",
                "routing",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["session_root"] == str(root.resolve())
    assert payload["entry_count"] == 1
    assert payload["shown_count"] == 1
    assert payload["kind_counts"]["routing"] == 1
    assert payload["surface_counts"]["read"] == 1
    assert (root / "state" / "oops.jsonl").exists()
    activity = content_activity.activity_events(root)
    assert activity[-1]["locator"] == "oops:session"
    assert activity[-1]["preferred_name"] == "oops:session"
    assert activity[-1]["follow_command"] == "gotta oops"


def test_oops_append_supports_from_file(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    payload_path = tmp_path / "oops.txt"
    initialize_session(root)
    payload_path.write_text(
        "Literal `gotta todo append` friction\nwith multiline context.\n",
        encoding="utf-8",
    )

    assert (
        oops.main(
            [
                "append",
                "--session",
                str(root),
                "--from-file",
                str(payload_path),
                "--surface",
                "todo",
                "--kind",
                "argument-shape-mismatch",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert "Literal `gotta todo append` friction" in payload["entries"][0]["message"]


def test_oops_append_resolves_from_file_relative_to_session(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    payload_path = root / "notes" / "oops.txt"
    initialize_session(root)
    payload_path.parent.mkdir(parents=True, exist_ok=True)
    payload_path.write_text("Session-relative oops payload\n", encoding="utf-8")
    monkeypatch.chdir(tmp_path)

    assert (
        oops.main(
            [
                "append",
                "--session",
                str(root),
                "--from-file",
                "notes/oops.txt",
                "--surface",
                "read",
                "--kind",
                "routing",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "Session-relative oops payload"


def test_oops_append_supports_from_file_dash_for_stdin(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    monkeypatch.setattr(oops.sys, "stdin", io.StringIO("stdin-backed oops payload\n"))

    assert (
        oops.main(
            [
                "append",
                "--session",
                str(root),
                "--from-file",
                "-",
                "--surface",
                "logs",
                "--kind",
                "input-shape",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "stdin-backed oops payload"


def test_oops_append_supports_stdin_flag(tmp_path: Path, monkeypatch, capsys) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    monkeypatch.setattr(oops.sys, "stdin", io.StringIO("explicit stdin oops payload\n"))

    assert (
        oops.main(
            [
                "append",
                "--session",
                str(root),
                "--stdin",
                "--surface",
                "logs",
                "--kind",
                "input-shape",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "explicit stdin oops payload"


def test_oops_write_only_inputs_imply_append_when_action_is_omitted(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    monkeypatch.setattr(oops.sys, "stdin", io.StringIO("implicit stdin oops payload\n"))

    assert (
        oops.main(
            [
                "--session",
                str(root),
                "--stdin",
                "--surface",
                "logs",
                "--kind",
                "input-shape",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "implicit stdin oops payload"


def test_oops_inline_prose_implies_append_when_action_is_omitted(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert (
        oops.main(
            [
                "--session",
                str(root),
                "--surface",
                "session-bootstrap",
                "minimal oops test",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "minimal oops test"
    assert payload["entries"][0]["surface"] == "session-bootstrap"


def test_oops_multiword_prose_implies_append_without_metadata(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert oops.main(["--session", str(root), "generic", "issue"]) == 0
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "generic issue"


def test_oops_single_unknown_token_without_write_intent_fails(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    capsys.readouterr()

    with pytest.raises(SystemExit) as exc:
        oops.main(["--session", str(root), "generic"])
    assert "unknown gotta oops action `generic`" in str(exc.value)

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 0


def test_oops_unknown_action_still_fails_instead_of_appending(
    tmp_path: Path, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    capsys.readouterr()

    with pytest.raises(SystemExit) as exc:
        oops.main(["--session", str(root), "smmary"])
    assert "unknown gotta oops action `smmary`" in str(exc.value)

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 0


@pytest.mark.parametrize("legacy_action", ["summary", "list"])
def test_oops_legacy_read_actions_redirect_to_show(
    tmp_path: Path, capsys, legacy_action: str
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    capsys.readouterr()

    with pytest.raises(SystemExit) as exc:
        oops.main([legacy_action, "--session", str(root)])
    assert (
        f"`gotta oops {legacy_action}` has been folded into `gotta oops show`"
        in str(exc.value)
    )


def test_oops_read_defaults_to_all_bound_actors_and_actor_filters_narrow(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert actor.main(["bind", "Claude", "Codex", "--session", str(root)]) == 0
    capsys.readouterr()
    claude = session_registry._resolve_bound_actor_name(root, "Claude")
    codex = session_registry._resolve_bound_actor_name(root, "Codex")

    monkeypatch.setenv(ACTOR_SPEAKER_ENV, codex)
    assert (
        oops.main(
            [
                "append",
                "codex-facing seam",
                "--session",
                str(root),
                "--actor",
                codex,
                "--surface",
                "session",
                "--kind",
                "routing",
            ]
        )
        == 0
    )
    capsys.readouterr()
    monkeypatch.setenv(ACTOR_SPEAKER_ENV, claude)
    assert (
        oops.main(
            [
                "append",
                "claude-facing seam",
                "--session",
                str(root),
                "--actor",
                claude,
                "--surface",
                "actor",
                "--kind",
                "contract",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["actor_count"] == 2
    assert payload["entry_count"] == 2
    assert payload["shown_count"] == 2
    assert set(payload["actors"]) == {codex, claude}
    assert set(payload["state_paths"]) == {codex, claude}

    assert {entry["actor"] for entry in payload["entries"]} == {codex, claude}

    assert (
        oops.main(
            ["show", "--session", str(root), "--actor", claude, "--output", "json"]
        )
        == 0
    )
    filtered = json.loads(capsys.readouterr().out)
    assert filtered["actor_count"] == 1
    assert filtered["actors"] == [claude]
    assert filtered["entry_count"] == 1


def test_oops_read_on_actor_root_defaults_to_session_wide(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert actor.main(["bind", "Claude", "Codex", "--session", str(root)]) == 0
    capsys.readouterr()
    claude = session_registry._resolve_bound_actor_name(root, "Claude")
    codex = session_registry._resolve_bound_actor_name(root, "Codex")
    actor_root = session_scope._session_dir(
        explicit_session=str(root), explicit_actor=claude
    )

    monkeypatch.setenv(ACTOR_SPEAKER_ENV, claude)
    assert (
        oops.main(
            [
                "append",
                "claude oops",
                "--session",
                str(actor_root),
                "--surface",
                "actor",
                "--kind",
                "contract",
            ]
        )
        == 0
    )
    capsys.readouterr()
    monkeypatch.setenv(ACTOR_SPEAKER_ENV, codex)
    assert (
        oops.main(
            [
                "append",
                "codex oops",
                "--session",
                str(root),
                "--actor",
                codex,
                "--surface",
                "session",
                "--kind",
                "routing",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(actor_root), "--output", "json"]) == 0
    listed = json.loads(capsys.readouterr().out)
    assert listed["actor_count"] == 2
    assert set(listed["actors"]) == {claude, codex}
    assert {entry["actor"] for entry in listed["entries"]} == {claude, codex}
    assert listed["shown_count"] == 2


def test_oops_append_writes_only_canonical_state(tmp_path: Path) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    friction.append_oops_record(
        root,
        message="append-hot-path entry",
        surface="read",
        kind="routing",
    )

    assert friction.oops_records(root)[0]["message"] == "append-hot-path entry"


def test_oops_concurrent_appends_only_touch_canonical_state(tmp_path: Path) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    total_threads = 4
    entries_per_thread = 25
    threads: list[threading.Thread] = []

    def append_entries(prefix: str) -> None:
        for index in range(entries_per_thread):
            friction.append_oops_record(
                root,
                message=f"{prefix}-{index}",
                surface="logs",
                kind="input-shape",
            )

    for thread_index in range(total_threads):
        thread = threading.Thread(
            target=append_entries,
            args=(f"t{thread_index}",),
            daemon=True,
        )
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join(timeout=5)
        assert not thread.is_alive()

    records = friction.oops_records(root)

    assert len(records) == total_threads * entries_per_thread
    for thread_index in range(total_threads):
        assert any(record["message"] == f"t{thread_index}-0" for record in records)
        assert any(
            record["message"] == f"t{thread_index}-{entries_per_thread - 1}"
            for record in records
        )


def test_oops_extend_supports_multiple_entries(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert (
        oops.main(
            [
                "extend",
                "First friction seam",
                "Second friction seam",
                "--session",
                str(root),
                "--surface",
                "read",
                "--kind",
                "routing",
            ]
        )
        == 0
    )
    capsys.readouterr()

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    messages = [entry["message"] for entry in payload["entries"]]
    assert "First friction seam" in messages
    assert "Second friction seam" in messages


def test_oops_bare_and_actor_reads_ignore_closed_stdin_without_mutating(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    assert actor.main(["bind", "Helper", "--session", str(root)]) == 0
    capsys.readouterr()
    helper = session_registry._resolve_bound_actor_name(root, "helper")

    assert oops.main(["append", "root friction", "--session", str(root)]) == 0
    capsys.readouterr()
    monkeypatch.setenv(ACTOR_SPEAKER_ENV, helper)
    assert (
        oops.main(
            ["append", "actor friction", "--session", str(root), "--actor", helper]
        )
        == 0
    )
    capsys.readouterr()

    root_before = (root / "state" / "oops.jsonl").read_text(encoding="utf-8")
    actor_root = session_scope._session_dir(
        explicit_session=str(root), explicit_actor=helper
    )
    actor_before = (actor_root / "state" / "oops.jsonl").read_text(encoding="utf-8")

    bare = _run_oops_subprocess(["--session", str(root)], closed_stdin=True)
    assert bare.returncode == 0
    assert "entries:" in bare.stdout
    assert "appended oops entry" not in bare.stdout
    assert (root / "state" / "oops.jsonl").read_text(encoding="utf-8") == root_before

    scoped = _run_oops_subprocess(["--session", str(actor_root)], closed_stdin=True)
    assert scoped.returncode == 0
    assert "entries:" in scoped.stdout
    assert "appended oops entry" not in scoped.stdout
    assert (actor_root / "state" / "oops.jsonl").read_text(
        encoding="utf-8"
    ) == actor_before


def test_oops_show_rejects_write_inputs_and_piped_stdin(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    payload_path = tmp_path / "friction.txt"
    initialize_session(root)
    assert oops.main(["append", "existing friction", "--session", str(root)]) == 0
    capsys.readouterr()
    payload_path.write_text("file-backed friction\n", encoding="utf-8")
    before = (root / "state" / "oops.jsonl").read_text(encoding="utf-8")

    piped = _run_oops_subprocess(
        ["show", "--session", str(root)],
        stdin_text="friction from stdin\n",
    )
    assert piped.returncode == 1
    assert "`gotta oops show` is read-only" in piped.stderr
    assert (root / "state" / "oops.jsonl").read_text(encoding="utf-8") == before

    with pytest.raises(SystemExit) as exc:
        oops.main(["show", "--session", str(root), "--stdin"])
    assert "`gotta oops show` is read-only" in str(exc.value)

    with pytest.raises(SystemExit) as exc:
        oops.main(["show", "--session", str(root), "--from-file", str(payload_path)])
    assert "`gotta oops show` is read-only" in str(exc.value)


def test_oops_real_piped_stdin_still_implies_append(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    capsys.readouterr()

    result = _run_oops_subprocess(
        ["--session", str(root)], stdin_text="piped friction\n"
    )
    assert result.returncode == 0
    assert "appended oops entry" in result.stdout

    assert oops.main(["show", "--session", str(root), "--output", "json"]) == 0
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 1
    assert payload["entries"][0]["message"] == "piped friction"


def test_oops_show_limit_zero_means_all_entries(tmp_path: Path, capsys) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert oops.main(["append", "first friction", "--session", str(root)]) == 0
    assert oops.main(["append", "second friction", "--session", str(root)]) == 0
    capsys.readouterr()

    assert (
        oops.main(["show", "--session", str(root), "--limit", "0", "--output", "json"])
        == 0
    )
    payload = json.loads(capsys.readouterr().out)
    assert payload["entry_count"] == 2
    assert payload["shown_count"] == 2


def test_foreign_writer_cannot_append_actor_oops(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)

    assert actor.main(["bind", "Claude", "--session", str(root)]) == 0
    capsys.readouterr()
    claude = session_registry._resolve_bound_actor_name(root, "Claude")
    monkeypatch.setenv(ACTOR_SPEAKER_ENV, "foreign-123")

    with pytest.raises(SystemExit) as excinfo:
        oops.main(
            [
                "append",
                "bad oops",
                "--session",
                str(root),
                "--actor",
                claude,
                "--surface",
                "read",
                "--kind",
                "routing",
            ]
        )

    assert "bind and launch a sibling actor" in str(excinfo.value)


def test_unbound_shell_cannot_append_actor_oops(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    assert actor.main(["bind", "Claude", "--session", str(root)]) == 0
    capsys.readouterr()
    claude = session_registry._resolve_bound_actor_name(root, "claude")
    monkeypatch.delenv(ACTOR_SPEAKER_ENV, raising=False)
    monkeypatch.delenv(content_env.ACTOR_ID_ENV, raising=False)
    monkeypatch.delenv(content_env.SESSION_ACTOR_ENV, raising=False)
    monkeypatch.setattr(
        content_context,
        "current_context_binding",
        lambda: type("Binding", (), {"binding_id": ""})(),
    )

    with pytest.raises(SystemExit) as excinfo:
        oops.cmd_oops(
            oops.build_parser().parse_args(
                ["append", "bad oops", "--session", str(root), "--actor", claude]
            )
        )

    assert "bind and launch a sibling actor" in str(excinfo.value)
    actor_root = session_scope._session_dir(
        explicit_session=str(root), explicit_actor=claude
    )
    assert friction.oops_records(actor_root) == []


def test_peer_actor_oops_preserves_peer_author(
    tmp_path: Path, monkeypatch, capsys
) -> None:
    root = tmp_path / "session-root"
    initialize_session(root)
    assert actor.main(["bind", "Claude", "Codex", "--session", str(root)]) == 0
    capsys.readouterr()
    claude = session_registry._resolve_bound_actor_name(root, "claude")
    codex = session_registry._resolve_bound_actor_name(root, "codex")
    monkeypatch.setenv(ACTOR_SPEAKER_ENV, codex)

    assert (
        oops.cmd_oops(
            oops.build_parser().parse_args(
                ["append", "peer oops", "--session", str(root), "--actor", claude]
            )
        )
        == 0
    )
    capsys.readouterr()

    actor_root = session_scope._session_dir(
        explicit_session=str(root), explicit_actor=claude
    )
    assert friction.oops_records(actor_root)[-1]["actor"] == codex



"""Lead surface for `gotta session`."""

from __future__ import annotations

import argparse
import json

from gotta.content.path import content_locator
from gotta.content.store import scan_content_store
from gotta.lead.aggregate import aggregate_lead_sources
from gotta.lead.edge import build_lead_edge_records
from gotta.lead.rank import edge_best_first_sort_key
from gotta.lead.resolve import resolve_lead_snapshots
from gotta.lead.snapshot import (
    snapshot_artifact_locator,
    snapshot_display_name,
    snapshot_last_fetched_at,
    snapshot_locator,
)

from .core import (
    LEADS_BEST_OVERALL_LIMIT,
    LEADS_PROVIDER_HIGHLIGHT_LIMIT,
    append_count_section,
    artifact_kind,
    compile_filter_pattern,
    filter_suffix,
    follow_command,
    lead_kind,
    lead_signal_labels,
    match_any,
    match_filter_text,
    no_leads_next_step,
    paging_summary_line,
    paginate_items,
    render_search_origins,
    resolved_visibility_metadata,
    stored_target_locators,
    top_count_records,
    topology_next_step,
    visibility_summary,
)
from .manifest import manifest_entries
from .parse import explicit_session_ref, require_started_session, session_dirs_for_read

_artifact_kind = artifact_kind
_append_count_section = append_count_section
_explicit_session_ref = explicit_session_ref
_manifest_entries = manifest_entries
_compile_filter_pattern = compile_filter_pattern
_filter_suffix = filter_suffix
_follow_command = follow_command
_lead_kind = lead_kind
_lead_signal_labels = lead_signal_labels
_match_any = match_any
_match_filter_text = match_filter_text
_no_leads_next_step = no_leads_next_step
_paging_summary_line = paging_summary_line
_paginate_items = paginate_items
_render_search_origins = render_search_origins
_resolved_visibility_metadata = resolved_visibility_metadata
_stored_target_locators = stored_target_locators
_top_count_records = top_count_records
_topology_next_step = topology_next_step
_visibility_summary = visibility_summary
_require_started_session = require_started_session
_session_dirs_for_read = session_dirs_for_read


def _leads_payload(
    dirs,
    *,
    target: str = "",
    filter_query: str = "",
    limit: int = 100,
    offset: int = 0,
    include_all: bool = False,
    session_ref: str = "",
) -> dict[str, object]:
    snapshots = scan_content_store(dirs.content_dir)
    manifest_entries = _manifest_entries(dirs)
    selected = resolve_lead_snapshots(target, snapshots, manifest_entries)
    selected_digests = {snapshot.digest for snapshot in selected}
    all_edge_records = build_lead_edge_records(
        snapshots,
        manifest_entries,
        classify_kind=_lead_kind,
    )
    edge_records = [
        edge
        for edge in all_edge_records
        if str(edge.get("sourceChecksum", "")) in selected_digests
    ]
    lead_sources = aggregate_lead_sources(edge_records)
    for lead in lead_sources:
        locator = str(lead.get("locator") or "").strip()
        if locator:
            lead["followCommand"] = _follow_command(locator, session_ref=session_ref)
    filter_text = _match_filter_text(filter_query)
    filter_pattern = _compile_filter_pattern(filter_text)
    if filter_pattern is not None:
        lead_sources = [
            lead
            for lead in lead_sources
            if _match_any(
                filter_pattern,
                lead.get("locator"),
                lead.get("followCommand"),
                lead.get("provider"),
                lead.get("kind"),
                lead.get("exampleRaw"),
                lead.get("contexts"),
                lead.get("relationKinds"),
                lead.get("artifactLocators"),
                lead.get("contentLocators"),
                lead.get("searchOrigins"),
            )
        ]
    paged_sources, paging = _paginate_items(
        lead_sources,
        limit=limit,
        offset=offset,
        include_all=include_all,
    )
    selected_lead_locators = {
        str(lead.get("locator") or "").strip()
        for lead in lead_sources
        if str(lead.get("locator") or "").strip()
    }
    selected_edges_by_checksum: dict[str, list[dict[str, object]]] = {}
    for edge in edge_records:
        if (
            selected_lead_locators
            and str(edge.get("targetLocator") or "").strip()
            not in selected_lead_locators
        ):
            continue
        selected_edges_by_checksum.setdefault(str(edge["sourceChecksum"]), []).append(
            edge
        )
    artifacts = []
    if lead_sources:
        for snapshot in selected:
            edges = sorted(
                selected_edges_by_checksum.get(snapshot.digest, []),
                key=edge_best_first_sort_key,
            )
            if filter_text and not edges:
                continue
            for edge in edges:
                target_locator = str(edge.get("targetLocator") or "").strip()
                if target_locator:
                    edge["followCommand"] = _follow_command(
                        target_locator,
                        session_ref=session_ref,
                    )
            artifacts.append(
                {
                    "checksum": snapshot.digest,
                    "preferredName": snapshot_display_name(snapshot),
                    "artifactKind": _artifact_kind(
                        snapshot.metadata.get("artifact_kind")
                    ),
                    "sourceLocator": snapshot_locator(snapshot),
                    "artifactLocator": snapshot_artifact_locator(snapshot),
                    "contentLocator": content_locator(snapshot.digest),
                    "lastFetchedAt": snapshot_last_fetched_at(snapshot),
                    "leadCount": len(edges),
                    "leads": edges[: max(limit, 0)],
                    **_resolved_visibility_metadata(
                        dict(snapshot.metadata),
                        provider=str(snapshot.metadata.get("plugin") or ""),
                        plugin=str(snapshot.metadata.get("plugin") or ""),
                        subcommand=str(snapshot.metadata.get("subcommand") or ""),
                        locator=str(snapshot_locator(snapshot)),
                    ),
                }
            )
    best_overall = lead_sources[:LEADS_BEST_OVERALL_LIMIT]
    best_locators = {str(item.get("locator") or "").strip() for item in best_overall}
    provider_highlights: list[dict[str, object]] = []
    highlighted_providers: set[str] = {
        str(item.get("provider") or "").strip()
        for item in best_overall
        if str(item.get("provider") or "").strip()
    }
    for lead in lead_sources:
        provider = str(lead.get("provider") or "").strip()
        locator = str(lead.get("locator") or "").strip()
        if (
            not provider
            or not locator
            or locator in best_locators
            or provider in highlighted_providers
        ):
            continue
        provider_highlights.append(lead)
        highlighted_providers.add(provider)
        if len(provider_highlights) >= LEADS_PROVIDER_HIGHLIGHT_LIMIT:
            break
    materialized_count = sum(
        1 for source in lead_sources if bool(source["materialized"])
    )
    discovery_count = sum(
        1 for item in artifacts if item.get("artifactKind") == "discovery"
    )
    evidence_count = sum(
        1 for item in artifacts if item.get("artifactKind") == "evidence"
    )
    empty = not artifacts and not lead_sources
    top_providers = _top_count_records(
        [str(source.get("provider") or "").strip() for source in lead_sources],
        key="provider",
    )
    top_relations = _top_count_records(
        [
            str(relation).strip()
            for source in lead_sources
            for relation in list(source.get("relationKinds") or [])
            if str(relation).strip()
        ],
        key="relation",
    )
    next_step = (
        _topology_next_step(
            discovery_count=discovery_count, evidence_count=evidence_count
        )
        if empty and not filter_text and not selected
        else _no_leads_next_step(has_artifacts=bool(selected))
        if not lead_sources and not filter_text and selected
        else (
            "No leads matched the current filter. Use `gotta session scan <query>` when you need corpus-wide search instead of field-level lead filtering."
            if not lead_sources and filter_text
            else ""
        )
    )
    return {
        "sessionDir": str(dirs.session_dir),
        "contentDir": str(dirs.content_dir),
        "target": target.strip(),
        "filter": filter_text,
        "limit": max(limit, 0),
        "artifactCount": len(artifacts),
        "discoveryArtifactCount": discovery_count,
        "evidenceArtifactCount": evidence_count,
        "topProviders": top_providers,
        "topRelations": top_relations,
        "leadCount": len(lead_sources),
        **paging,
        "materializedLeadCount": materialized_count,
        "unmaterializedLeadCount": len(lead_sources) - materialized_count,
        "leadSources": paged_sources,
        "bestOverall": best_overall,
        "providerHighlights": provider_highlights,
        "artifacts": artifacts,
        "empty": empty,
        "nextStep": next_step,
    }


def cmd_leads(args: argparse.Namespace) -> int:
    dirs = _session_dirs_for_read(args)
    _require_started_session(dirs)
    session_ref = _explicit_session_ref(args)
    payload = _leads_payload(
        dirs,
        target=args.target or "",
        filter_query=str(getattr(args, "filter", "") or ""),
        session_ref=session_ref,
        limit=max(args.limit, 0),
        offset=max(args.offset, 0),
        include_all=bool(args.all),
    )
    if args.output == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    print(f"session: {payload['sessionDir']}")
    print(f"target: {payload['target'] or '(session-wide)'}")
    print(
        "artifacts: "
        f"{payload['artifactCount']} "
        f"(discovery {payload['discoveryArtifactCount']}, "
        f"evidence {payload['evidenceArtifactCount']})"
    )
    print(
        "leads: "
        f"{payload['leadCount']} total (showing {payload['shownCount']}; "
        f"materialized {payload['materializedLeadCount']}, "
        f"unmaterialized {payload['unmaterializedLeadCount']})"
        f"{_filter_suffix(payload.get('filter'))}"
    )
    print(
        _paging_summary_line(
            label="page",
            total_count=int(payload["totalCount"]),
            shown_count=int(payload["shownCount"]),
            offset=int(payload["offset"]),
            next_offset=(
                int(payload["nextOffset"])
                if payload.get("nextOffset") is not None
                else None
            ),
        )
    )
    if int(payload["shownCount"]) == 0 and int(payload["totalCount"]) > 0:
        print("page: no results in this page window")
    top_providers_lines: list[str] = []
    _append_count_section(
        top_providers_lines,
        heading="top providers",
        records=list(payload.get("topProviders") or []),
        key="provider",
    )
    if top_providers_lines:
        print("\n".join(top_providers_lines))
    top_relation_lines: list[str] = []
    _append_count_section(
        top_relation_lines,
        heading="top relations",
        records=list(payload.get("topRelations") or []),
        key="relation",
    )
    if top_relation_lines:
        print("\n".join(top_relation_lines))
    if payload["nextStep"]:
        print(f"next: {payload['nextStep']}")
    if payload["bestOverall"]:
        print("best leads:")
        for lead in payload["bestOverall"]:
            relation = ", ".join(
                str(value) for value in lead.get("relationKinds") or [] if str(value)
            )
            print(
                f"  - [{'; '.join(_lead_signal_labels(lead, aggregated=True))}] "
                f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
            )
            visibility = _visibility_summary(lead)
            if visibility:
                print(f"    visibility: {visibility}")
            print(f"    follow: `{lead['followCommand']}`")
            stored_targets = _stored_target_locators(lead)
            if stored_targets:
                print(
                    "    stored_target: "
                    + ", ".join(f"`{value}`" for value in stored_targets)
                )
            search_origins = _render_search_origins(lead)
            if search_origins:
                print(f"    search_origin: {search_origins}")
            contexts = [
                str(value) for value in lead.get("contexts") or [] if str(value)
            ]
            if contexts:
                print(f"    context: {contexts[0]}")
    if payload["providerHighlights"]:
        print("provider highlights:")
        for lead in payload["providerHighlights"]:
            relation = ", ".join(
                str(value) for value in lead.get("relationKinds") or [] if str(value)
            )
            print(
                f"  - [{'; '.join(_lead_signal_labels(lead, aggregated=True))}] "
                f"{lead['locator']} ({lead['provider']}, {relation or 'lead'})"
            )
            print(f"    follow: `{lead['followCommand']}`")
    if payload["artifacts"] and (payload["target"] or payload["artifactCount"] == 1):
        print("source context:")
        for artifact in payload["artifacts"]:
            print(f"- {artifact['preferredName']} ({str(artifact['checksum'])[:12]})")
            print(f"  source: `{artifact['sourceLocator'] or 'unknown'}`")
            if artifact.get("artifactKind"):
                print(f"  artifact_kind: {artifact['artifactKind']}")
            artifact_visibility = _visibility_summary(artifact)
            if artifact_visibility:
                print(f"  visibility: {artifact_visibility}")
            print(
                f"  stored: `{artifact['artifactLocator']}`, `{artifact['contentLocator']}`"
            )
            if artifact["lastFetchedAt"]:
                print(f"  fetched: {artifact['lastFetchedAt']}")
            if not artifact["leads"]:
                print("  leads: none")
                continue
            for lead in artifact["leads"]:
                print(
                    f"  - [{'; '.join(_lead_signal_labels(lead, aggregated=False))}] "
                    f"{lead['targetLocator']} ({lead['provider']}, {lead['relation']})"
                )
                visibility = _visibility_summary(lead)
                if visibility:
                    print(f"    visibility: {visibility}")
                print(f"    follow: `{lead['followCommand']}`")
                stored_targets = _stored_target_locators(lead)
                if stored_targets:
                    print(
                        "    stored_target: "
                        + ", ".join(f"`{value}`" for value in stored_targets)
                    )
                if bool(lead.get("sourceSearchLike")):
                    provider = str(lead.get("sourceProvider") or "unknown")
                    subcommand = str(lead.get("sourceSubcommand") or "search")
                    rank = int(lead.get("sourceRank") or 0)
                    origin = f"{provider}/{subcommand}"
                    if rank > 0:
                        origin += f" #{rank}"
                    print(f"    source_origin: {origin}")
                contexts = [
                    str(value) for value in lead.get("contexts") or [] if str(value)
                ]
                if contexts:
                    print(f"    context: {contexts[0]}")
    return 0


leads_payload = _leads_payload



"""Build lead edges from session content snapshots."""

from __future__ import annotations

from gotta.content.model import ContentSnapshot
from gotta.content.path import content_locator
from gotta.source import best_visibility_metadata, classify_visibility_metadata

from .cache import lead_mentions_for_snapshot
from .rank import edge_best_first_sort_key, is_first_party_target
from .snapshot import (
    snapshot_artifact_locator,
    snapshot_display_name,
    snapshot_is_search_like,
    snapshot_locator,
    snapshot_provider,
    snapshot_sort_key,
    snapshot_subcommand,
)


def materialized_source_index(
    manifest_entries: list[dict[str, object]],
    snapshot_by_digest: dict[str, ContentSnapshot],
) -> dict[str, set[str]]:
    index: dict[str, set[str]] = {}
    for entry in manifest_entries:
        locator = str(
            entry.get("canonical_locator", "") or entry.get("locator", "")
        ).strip()
        checksum = str(entry.get("checksum", "")).strip()
        if not locator or not checksum or checksum not in snapshot_by_digest:
            continue
        index.setdefault(locator, set()).add(checksum)
    return index


def materialized_target_locators(
    locator: str,
    source_index: dict[str, set[str]],
    snapshot_by_digest: dict[str, ContentSnapshot],
) -> tuple[list[str], list[str]]:
    digests = sorted(source_index.get(locator, set()))
    artifact_locators = [
        snapshot_artifact_locator(snapshot_by_digest[digest])
        for digest in digests
        if digest in snapshot_by_digest
    ]
    content_locators = [content_locator(digest) for digest in digests]
    return artifact_locators, content_locators


def materialized_target_visibility(
    locator: str,
    source_index: dict[str, set[str]],
    snapshot_by_digest: dict[str, ContentSnapshot],
) -> dict[str, object]:
    return best_visibility_metadata(
        *[
            snapshot_by_digest[digest].metadata
            for digest in sorted(source_index.get(locator, set()))
            if digest in snapshot_by_digest
        ]
    )


def build_lead_edge_records(
    snapshots: list[ContentSnapshot],
    manifest_entries: list[dict[str, object]],
    *,
    classify_kind,
) -> list[dict[str, object]]:
    snapshot_by_digest = {snapshot.digest: snapshot for snapshot in snapshots}
    source_index = materialized_source_index(manifest_entries, snapshot_by_digest)
    rendered: list[dict[str, object]] = []
    for snapshot in sorted(snapshots, key=snapshot_sort_key, reverse=True):
        source_locator = snapshot_locator(snapshot)
        source_search_like = snapshot_is_search_like(snapshot)
        source_provider = snapshot_provider(snapshot)
        source_subcommand = snapshot_subcommand(snapshot)
        edge_state: dict[str, dict[str, object]] = {}
        for mention in lead_mentions_for_snapshot(snapshot):
            target_locator = mention.canonical_locator.strip()
            if not target_locator or target_locator == source_locator:
                continue
            state = edge_state.setdefault(
                target_locator,
                {
                    "sourceChecksum": snapshot.digest,
                    "sourceLocator": source_locator or content_locator(snapshot.digest),
                    "sourceArtifactLocator": snapshot_artifact_locator(snapshot),
                    "sourceContentLocator": content_locator(snapshot.digest),
                    "sourcePreferredName": snapshot_display_name(snapshot),
                    "targetLocator": target_locator,
                    "provider": mention.provider,
                    "kind": classify_kind(target_locator, mention.provider),
                    "relation": mention.relation,
                    "followCommand": mention.follow_command,
                    "rawExamples": [],
                    "contexts": [],
                    "occurrenceCount": 0,
                    "sourceSearchLike": source_search_like,
                    "sourceProvider": source_provider,
                    "sourceSubcommand": source_subcommand,
                    "sourceRank": mention.ordinal if source_search_like else 0,
                },
            )
            state["occurrenceCount"] = int(state["occurrenceCount"]) + 1
            mention_ordinal = int(mention.ordinal or 0) if source_search_like else 0
            if mention_ordinal and (
                not int(state.get("sourceRank") or 0)
                or mention_ordinal < int(state.get("sourceRank") or 0)
            ):
                state["sourceRank"] = mention_ordinal
            if (
                mention.raw not in state["rawExamples"]
                and len(state["rawExamples"]) < 3
            ):
                state["rawExamples"].append(mention.raw)
            if (
                mention.snippet
                and mention.snippet not in state["contexts"]
                and len(state["contexts"]) < 3
            ):
                state["contexts"].append(mention.snippet)
        for state in edge_state.values():
            artifact_locators, content_locators = materialized_target_locators(
                str(state["targetLocator"]),
                source_index,
                snapshot_by_digest,
            )
            target_visibility = best_visibility_metadata(
                classify_visibility_metadata(
                    {},
                    provider=str(state["provider"]),
                    locator=str(state["targetLocator"]),
                ),
                materialized_target_visibility(
                    str(state["targetLocator"]),
                    source_index,
                    snapshot_by_digest,
                ),
            )
            state["materialized"] = bool(artifact_locators or content_locators)
            state["targetArtifactLocators"] = artifact_locators
            state["targetContentLocators"] = content_locators
            state["firstParty"] = is_first_party_target(
                provider=str(state["provider"]),
                kind=str(state["kind"]),
            )
            state["searchSeed"] = str(state["kind"]).endswith("-search")
            state.update(target_visibility)
            rendered.append(state)
    return sorted(
        rendered,
        key=lambda item: (
            str(item["sourcePreferredName"]).casefold(),
            *edge_best_first_sort_key(item),
        ),
    )



"""Top-level `gotta search` routing."""

from __future__ import annotations

from dataclasses import dataclass
import shlex

from gotta.builtin import get_plugin
from gotta.resolve.read import discover_plugin_route


SPECIALIZED_PROVIDER_COMMANDS: dict[tuple[str, str], str] = {
    ("jira", "jql"): "jira jql",
    ("confluence", "cql"): "confluence cql",
    ("granola", "search-transcript"): "granola search-transcript",
}

READ_LIKE_PROVIDER_COMMANDS = {"get", "transcript"}
READ_LIKE_TARGET_PREFIXES = ("http://", "https://", "artifact:", "content:")


@dataclass(frozen=True, slots=True)
class SearchRoute:
    provider: str
    provider_argv: list[str]
    target: str


class SearchRouteError(ValueError):
    """Raised when a top-level search target cannot be routed cleanly."""


def _first_word(text: str) -> str:
    if not text.strip():
        return ""
    try:
        parts = shlex.split(text)
    except ValueError:
        parts = text.split()
    return parts[0].strip() if parts else ""


def _raw_tail(argv: list[str]) -> tuple[str, str, str]:
    if not argv:
        raise SearchRouteError(
            "search target must start with <provider>:<query>; use `gotta search jira:Architecture`"
        )
    target = str(argv[0] or "").strip()
    if not target:
        raise SearchRouteError(
            "search target must start with <provider>:<query>; use `gotta search jira:Architecture`"
        )
    if target.startswith("-"):
        raise SearchRouteError(
            "search target must start with <provider>:<query>; flags belong after the provider-qualified target"
        )
    if target.startswith(READ_LIKE_TARGET_PREFIXES):
        raise SearchRouteError(
            f"read-like targets belong on `gotta read`; use `gotta read {shlex.join(argv)}`"
        )
    provider, separator, tail = target.partition(":")
    provider = provider.strip()
    if not separator or not provider or not tail.strip():
        raise SearchRouteError(
            "search target must start with <provider>:<query>; use `gotta search jira:Architecture`"
        )
    extra_terms = [
        str(raw_part or "").strip()
        for raw_part in argv[1:]
        if str(raw_part or "").strip()
    ]
    if extra_terms:
        raise SearchRouteError(
            "top-level `gotta search` takes exactly one provider-qualified plain-text query string; "
            f"quote the full `<provider>:<query>` argument or use `gotta {provider} search ...` "
            "for structured flags"
        )
    raw_tail = tail.strip()
    if not raw_tail:
        raise SearchRouteError(
            "search target must include a query seed after the provider; use `gotta search jira:Architecture`"
        )
    return provider, raw_tail, target


def _canonical_read_locator(
    provider: str,
    provider_argv: list[str],
    *,
    fallback: str,
) -> str:
    spec = get_plugin(provider)
    if spec and spec.canonical_locator is not None:
        locator = str(spec.canonical_locator(provider_argv) or "").strip()
        if locator:
            return locator
    return fallback


def _tail_after_first_word(text: str) -> str:
    stripped = text.strip()
    if not stripped:
        return ""
    try:
        parts = shlex.split(stripped)
    except ValueError:
        _first, _separator, rest = stripped.partition(" ")
        return rest.strip()
    if len(parts) < 2:
        return ""
    return " ".join(parts[1:]).strip()


def _discover_read_locator(provider: str, candidate: str) -> str:
    routed = discover_plugin_route(candidate)
    if routed is None:
        return ""
    routed_provider, provider_argv = routed
    if routed_provider != provider or not provider_argv:
        return ""
    if candidate.startswith(READ_LIKE_TARGET_PREFIXES):
        return _canonical_read_locator(provider, provider_argv, fallback=candidate)
    if provider_argv[0] not in READ_LIKE_PROVIDER_COMMANDS:
        return ""
    return _canonical_read_locator(provider, provider_argv, fallback=candidate)


def _canonical_read_redirect(provider: str, raw_tail: str) -> str:
    first = _first_word(raw_tail)
    subject = _tail_after_first_word(raw_tail)
    if not first or not subject:
        return ""
    candidates: list[str] = []
    if first == "get":
        if subject.startswith(READ_LIKE_TARGET_PREFIXES):
            candidates.append(subject)
        if provider == "github":
            normalized = subject.strip()
            if normalized.startswith("github.com/"):
                candidates.append(f"https://{normalized}")
            elif "/" in normalized and " " not in normalized:
                candidates.append(f"https://github.com/{normalized.lstrip('/')}")
        if provider == "slack" and " " not in subject:
            if ":" in subject:
                channel_id, thread_ref = subject.split(":", 1)
                if channel_id and thread_ref:
                    candidates.append(f"slack:thread:{channel_id}:{thread_ref}")
            else:
                candidates.append(f"slack:channel:{subject}")
        candidates.append(f"{provider}:{subject}")
    elif first == "transcript":
        candidates.append(f"{provider}:transcript {subject}")
    seen: set[str] = set()
    for candidate in candidates:
        if candidate in seen:
            continue
        seen.add(candidate)
        locator = _discover_read_locator(provider, candidate)
        if locator:
            return locator
    return ""


def resolve_search_route(argv: list[str]) -> SearchRoute:
    provider, raw_tail, target = _raw_tail(argv)
    spec = get_plugin(provider)
    if (
        spec is None
        or spec.route_target is None
        or provider in {"read", "session", "search"}
    ):
        raise SearchRouteError(
            f"unknown search provider `{provider}`; use one of the routed provider plugins"
        )
    first = _first_word(raw_tail)
    specialized = SPECIALIZED_PROVIDER_COMMANDS.get((provider, first))
    if specialized:
        raise SearchRouteError(
            f"`gotta search` is for plain-text search; use `gotta {specialized}` for `{provider}:{first}`"
        )
    if first in READ_LIKE_PROVIDER_COMMANDS:
        read_target = _canonical_read_redirect(provider, raw_tail)
        if read_target:
            suggestion = shlex.join(["gotta", "read", read_target])
        else:
            suggestion = shlex.join(["gotta", provider, raw_tail])
        raise SearchRouteError(
            f"read-like provider targets belong on `gotta read`; use `{suggestion}`"
        )
    explicit_alias = first == "search"
    query = _tail_after_first_word(raw_tail) if explicit_alias else raw_tail
    query = query.strip()
    if not query:
        raise SearchRouteError(
            "search target must include a query seed after the provider; use `gotta search jira:Architecture`"
        )
    return SearchRoute(
        provider=provider,
        provider_argv=["search", query],
        target=target,
    )



"""Captured stream helpers for dispatch."""

from __future__ import annotations

from contextlib import contextmanager
import io
import sys
from typing import Any


class _CapturedBuffer(io.RawIOBase):
    def __init__(self, backing: io.BytesIO, passthrough: Any | None = None) -> None:
        self._backing = backing
        self._passthrough = passthrough

    def write(self, data: bytes | bytearray) -> int:
        payload = bytes(data)
        self._backing.write(payload)
        if self._passthrough is not None:
            self._passthrough.write(payload)
            self._passthrough.flush()
        return len(payload)

    def flush(self) -> None:
        if self._passthrough is not None:
            self._passthrough.flush()


class CapturedStream(io.TextIOBase):
    """Capture text and binary writes while optionally mirroring to the real stream."""

    def __init__(
        self,
        stream: Any,
        *,
        preserve_tty: bool = False,
        passthrough: bool = False,
    ) -> None:
        self._stream = stream
        self._encoding = getattr(stream, "encoding", None) or "utf-8"
        self._isatty = (
            bool(stream.isatty())
            if preserve_tty and hasattr(stream, "isatty")
            else False
        )
        self._buffer = io.BytesIO()
        passthrough_buffer = getattr(stream, "buffer", None) if passthrough else None
        self.buffer = _CapturedBuffer(self._buffer, passthrough_buffer)
        self._passthrough = stream if passthrough else None

    @property
    def encoding(self) -> str:
        return self._encoding

    def write(self, data: str) -> int:
        payload = data.encode(self._encoding)
        self._buffer.write(payload)
        if self._passthrough is not None:
            self._passthrough.write(data)
            self._passthrough.flush()
        return len(data)

    def flush(self) -> None:
        if self._passthrough is not None:
            self._passthrough.flush()

    def isatty(self) -> bool:
        return self._isatty

    def getvalue(self) -> bytes:
        return self._buffer.getvalue()


@contextmanager
def capture_stdout(*, preserve_tty: bool = False, passthrough: bool = False) -> Any:
    original_stdout = sys.stdout
    capture = CapturedStream(
        original_stdout,
        preserve_tty=preserve_tty,
        passthrough=passthrough,
    )
    sys.stdout = capture
    try:
        yield capture
    finally:
        sys.stdout = original_stdout


@contextmanager
def capture_stderr(*, preserve_tty: bool = False, passthrough: bool = False) -> Any:
    original_stderr = sys.stderr
    capture = CapturedStream(
        original_stderr,
        preserve_tty=preserve_tty,
        passthrough=passthrough,
    )
    sys.stderr = capture
    try:
        yield capture
    finally:
        sys.stderr = original_stderr


def _emit_captured(data: bytes) -> None:
    if not data:
        return
    buffer = getattr(sys.stdout, "buffer", None)
    if buffer is not None:
        buffer.write(data)
        sys.stdout.flush()
        return
    sys.stdout.write(data.decode("utf-8", errors="replace"))
    sys.stdout.flush()


def _emit_captured_stderr(data: bytes) -> None:
    if not data:
        return
    buffer = getattr(sys.stderr, "buffer", None)
    if buffer is not None:
        buffer.write(data)
        sys.stderr.flush()
        return
    sys.stderr.write(data.decode("utf-8", errors="replace"))
    sys.stderr.flush()



"""Top-level session-rooted friction capture."""

from __future__ import annotations

import argparse
import json
import sys
from pathlib import Path

from gotta.actor import require_writer, session_actor, writer_name, writer_role
from gotta.content.context import stdin_has_meaningful_text
from gotta.content.model import CommonOptions, ContentError
from gotta.content.path import session_relative_path
from gotta.content.scope import resolve_dirs, session_is_initialized
from gotta.helptext import format_long_help, is_long_help_request
from gotta import topology
from gotta.friction import (
    append_oops_record,
    filtered_oops_records,
    oops_log_path,
    oops_records,
    oops_summary,
)
from gotta.session import charter as session_charter
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope

OOPS_ACTIONS = {"show", "append", "extend"}
OOPS_LEGACY_READ_ACTIONS = {"list", "summary"}
_APPEND_METADATA_FLAGS = (
    "--surface",
    "--command",
    "--kind",
    "--affordance",
    "--workaround",
    "--severity",
    "--reproducibility",
    "--resolution-state",
)


def _has_explicit_write_source(argv: list[str]) -> bool:
    return any(
        token == "--stdin" or token == "--from-file" or token.startswith("--from-file=")
        for token in argv
    )


def _argv_has_append_metadata(argv: list[str]) -> bool:
    for index, token in enumerate(argv):
        if any(token.startswith(f"{flag}=") for flag in _APPEND_METADATA_FLAGS):
            return True
        if token in _APPEND_METADATA_FLAGS and index + 1 < len(argv):
            return True
    return False


def _has_append_metadata(args: argparse.Namespace) -> bool:
    return any(
        (
            str(args.surface or "").strip(),
            str(args.command or "").strip(),
            str(args.kind or "").strip(),
            str(args.affordance or "").strip(),
            str(args.workaround or "").strip(),
            str(args.severity or "").strip(),
            str(args.reproducibility or "").strip() != "unknown",
            str(args.resolution_state or "").strip() != "open",
        )
    )


def _read_text_source(
    *,
    session_root: Path,
    inline: str | None,
    from_file: str | None,
    use_stdin: bool,
    input_name: str,
) -> str:
    used = int(bool(inline)) + int(bool(from_file)) + int(bool(use_stdin))
    if used > 1:
        raise SystemExit(f"use only one {input_name} source")
    if from_file:
        if from_file == "-":
            return sys.stdin.read()
        return session_relative_path(session_root, from_file).read_text(
            encoding="utf-8"
        )
    if use_stdin:
        return sys.stdin.read()
    if inline is not None:
        return inline
    if stdin_has_meaningful_text():
        return sys.stdin.read()
    raise SystemExit(
        f"missing {input_name}; pass inline text, use --stdin, use --from-file, or pipe stdin"
    )


def _read_text_items_source(
    *,
    session_root: Path,
    inline_items: list[str],
    from_file: str | None,
    use_stdin: bool,
    input_name: str,
) -> list[str]:
    used = int(bool(inline_items)) + int(bool(from_file)) + int(bool(use_stdin))
    if used > 1:
        raise SystemExit(f"use only one {input_name} source")
    if from_file:
        if from_file == "-":
            raw = sys.stdin.read()
        else:
            raw = session_relative_path(session_root, from_file).read_text(
                encoding="utf-8"
            )
    elif use_stdin:
        raw = sys.stdin.read()
    elif inline_items:
        raw = "\n".join(inline_items)
    elif stdin_has_meaningful_text():
        raw = sys.stdin.read()
    else:
        raise SystemExit(
            f"missing {input_name}; pass one or more inline entries, use --stdin, use --from-file, or pipe stdin"
        )
    items: list[str] = []
    for line in raw.replace("\r\n", "\n").replace("\r", "\n").splitlines():
        normalized = line.strip()
        if not normalized:
            continue
        items.append(normalized)
    if not items:
        raise SystemExit(f"missing {input_name}")
    return items


def _normalize_text(text: str, *, input_name: str) -> str:
    normalized = text.replace("\r\n", "\n").replace("\r", "\n").strip("\n")
    if not normalized.strip():
        raise SystemExit(f"missing {input_name}")
    return normalized


def _resolve_session_root(*, explicit_session: str | None) -> Path:
    session_raw = explicit_session
    try:
        dirs = resolve_dirs(
            CommonOptions(
                session_dir=session_raw,
            ),
            create=False,
        )
    except ContentError as exc:
        raise SystemExit(str(exc)) from exc
    session_dir = dirs.session_dir
    if not session_is_initialized(session_dir):
        raise SystemExit(
            "start or bind a session first with `gotta ...`. Stable interactive "
            "contexts adopt and scaffold their deterministic session on first "
            'session-aware use. Use `gotta session init --session "$WS"` only '
            "when you intentionally want to scaffold one exact root."
        )
    return session_dir


def build_parser(command_name: str = "gotta oops") -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=command_name,
        description=(
            "Show or record durable session speed bumps, including suspected gotta bugs. "
            "Bare `gotta oops` shows the friction ledger across the active session scope. "
            "Bare multiword prose, real piped stdin, `--stdin`, and `--from-file` imply "
            "`append` when no read action is named. "
            "Use `gotta oops show` for explicit reads and `gotta oops append ...` when you "
            "want write intent to be unambiguous."
        ),
    )
    parser.add_argument("action", nargs="?")
    parser.add_argument("value", nargs="*")
    session_charter.add_target_args(parser)
    parser.add_argument(
        "--from-file",
        help="read friction text from a UTF-8 file instead of inline text; use '-' for stdin",
    )
    parser.add_argument(
        "--stdin",
        dest="use_stdin",
        action="store_true",
        help="read friction text from stdin explicitly",
    )
    parser.add_argument("--surface", help="surface where the friction occurred")
    parser.add_argument("--command", help="exact command or subcommand involved")
    parser.add_argument("--kind", help="blocker kind, e.g. auth, output, routing")
    parser.add_argument("--affordance", help="implicated affordance or contract seam")
    parser.add_argument("--workaround", help="workaround used to continue")
    parser.add_argument(
        "--severity",
        choices=["low", "medium", "high", "critical"],
        default="",
    )
    parser.add_argument(
        "--reproducibility",
        choices=["unknown", "intermittent", "reliable"],
        default="unknown",
    )
    parser.add_argument(
        "--resolution-state",
        choices=["open", "accepted", "resolved"],
        default="open",
    )
    parser.add_argument("--output", choices=["json", "text"], default="text")
    parser.add_argument("--limit", type=int, default=20)
    return parser


def session_access_mode(argv: list[str]) -> str:
    positionals = session_charter.argv_positionals(
        argv,
        valued_flags=(
            "--session",
            "--actor",
            "--from-file",
            "--surface",
            "--command",
            "--kind",
            "--affordance",
            "--workaround",
            "--severity",
            "--reproducibility",
            "--resolution-state",
            "--output",
            "--limit",
        ),
    )
    action = positionals[0] if positionals else ""
    if action in {"append", "extend"}:
        return "write"
    if action in {"show", "list", "summary"}:
        return "read"
    if not action:
        if _has_explicit_write_source(argv) or stdin_has_meaningful_text():
            return "write"
        return "read"
    if (
        _has_explicit_write_source(argv)
        or stdin_has_meaningful_text()
        or len(positionals) > 1
        or (_argv_has_append_metadata(argv) and bool(positionals))
    ):
        return "write"
    return "read"


def _aggregate_oops_records(
    work_dir: Path,
    *,
    actor_name: str | None = None,
    surface: str = "",
    command: str = "",
    kind: str = "",
    severity: str = "",
) -> tuple[list[str], list[dict[str, object]]]:
    actor_ids = list(session_scope._target_actor_ids(work_dir, actor_name))
    records: list[dict[str, object]] = []
    for current_actor in actor_ids:
        actor_root = session_registry._actor_session_dir(work_dir, current_actor)
        for record in filtered_oops_records(
            oops_records(actor_root),
            surface=surface,
            command=command,
            kind=kind,
            severity=severity,
        ):
            if (
                actor_root.resolve().parent.name == "actors"
                and writer_role(
                    actor_root, current_actor, writer=str(record.get("actor") or "")
                )
                == "foreign"
            ):
                continue
            payload = dict(record)
            payload.setdefault("actor", current_actor)
            payload.setdefault(
                "label", session_registry._actor_label(current_actor, work_dir=work_dir)
            )
            records.append(payload)
    records = sorted(
        records, key=lambda item: str(item.get("timestamp") or ""), reverse=True
    )
    return actor_ids, records


def _explicit_read_redirect(action_token: str) -> SystemExit:
    return SystemExit(
        f"`gotta oops {action_token}` has been folded into `gotta oops show`; "
        "use `gotta oops` or `gotta oops show`"
    )


def _read_write_conflict() -> SystemExit:
    return SystemExit(
        "`gotta oops show` is read-only; remove stdin/file input and inline prose, "
        "or use `gotta oops append ...` to write"
    )


def _unknown_action(action_token: str) -> SystemExit:
    return SystemExit(
        f"unknown gotta oops action `{action_token}`; use `show`, `append`, or `extend`, "
        "or pass multiword prose / real stdin to imply append"
    )


def _limited_records(
    records: list[dict[str, object]], *, limit: int
) -> list[dict[str, object]]:
    bounded = max(limit, 0)
    if bounded == 0:
        return list(records)
    return records[:bounded]


def _read_payload(
    session_dir: Path,
    actor_ids: list[str],
    records: list[dict[str, object]],
    *,
    explicit_actor: str | None,
    scoped_actor: str | None,
    limit: int,
) -> dict[str, object]:
    limited = _limited_records(records, limit=limit)
    payload: dict[str, object] = {
        "session_root": str(session_dir),
        "actor_count": len(actor_ids),
        "actors": actor_ids,
        "shown_count": len(limited),
        "entries": limited,
        **oops_summary(records),
    }
    if explicit_actor or scoped_actor:
        payload["state_path"] = str(oops_log_path(session_dir))
        payload["locator"] = session_charter._native_surface_locator(
            "oops",
            actor_name=scoped_actor or session_actor(session_dir),
        )
        payload["follow_command"] = session_charter._native_surface_follow_command(
            "oops",
            actor_name=scoped_actor or session_actor(session_dir),
        )
    else:
        payload["state_paths"] = {
            actor: str(
                oops_log_path(session_registry._actor_session_dir(session_dir, actor))
            )
            for actor in actor_ids
        }
        payload["locators"] = {
            actor: session_charter._native_surface_locator("oops", actor_name=actor)
            for actor in actor_ids
        }
        payload["follow_commands"] = {
            actor: session_charter._native_surface_follow_command(
                "oops", actor_name=actor
            )
            for actor in actor_ids
        }
    return payload


def _is_exact_session_root(work_dir: Path) -> bool:
    resolved = work_dir.resolve()
    return (
        resolved.parent.name != "actors"
        and topology.parse_grouped_session_root(resolved) is None
        and topology.parse_shared_session_root(resolved) is None
        and session_is_initialized(resolved)
    )


def _resolved_action(args: argparse.Namespace) -> tuple[str, list[str]]:
    values = list(args.value or [])
    action_token = str(args.action or "").strip()
    explicit_write_source = bool(args.from_file or args.use_stdin)
    piped_text = stdin_has_meaningful_text()
    has_inline_prose = bool(values)
    has_append_metadata = _has_append_metadata(args) and bool(action_token or values)
    if action_token in OOPS_LEGACY_READ_ACTIONS:
        raise _explicit_read_redirect(action_token)
    if action_token in {"show"}:
        if explicit_write_source or piped_text or values:
            raise _read_write_conflict()
        return action_token, values
    if action_token in OOPS_ACTIONS:
        return action_token, values
    if action_token and (
        explicit_write_source or has_append_metadata or has_inline_prose
    ):
        values = [action_token, *values]
        return "append", values
    if not action_token and (explicit_write_source or piped_text):
        return "append", values
    if not action_token:
        return "show", values
    raise _unknown_action(action_token)


def cmd_oops(args: argparse.Namespace) -> int:
    action, values = _resolved_action(args)
    if action == "append":
        session_dir = session_scope._session_dir(
            explicit_session=getattr(args, "session", None),
            explicit_actor=getattr(args, "actor", None),
        )
        writer = writer_name()
        actor_branch = session_dir.resolve().parent.name == "actors"
        target_actor = session_actor(session_dir) if actor_branch else ""
        if target_actor:
            require_writer(
                session_dir,
                target_actor,
                writer=writer,
                action="write into this actor branch",
            )
        payload = _read_text_source(
            session_root=session_dir,
            inline=(" ".join(values) if values else None),
            from_file=args.from_file,
            use_stdin=args.use_stdin,
            input_name="oops entry text",
        )
        append_oops_record(
            session_dir,
            message=_normalize_text(payload, input_name="oops entry text"),
            actor=writer,
            surface=args.surface or "",
            command=args.command or "",
            kind=args.kind or "",
            affordance=args.affordance or "",
            workaround=args.workaround or "",
            severity=args.severity or "medium",
            reproducibility=args.reproducibility,
            resolution_state=args.resolution_state,
        )
        print("appended oops entry")
        return 0
    if action == "extend":
        session_dir = session_scope._session_dir(
            explicit_session=getattr(args, "session", None),
            explicit_actor=getattr(args, "actor", None),
        )
        writer = writer_name()
        actor_branch = session_dir.resolve().parent.name == "actors"
        target_actor = session_actor(session_dir) if actor_branch else ""
        if target_actor:
            require_writer(
                session_dir,
                target_actor,
                writer=writer,
                action="write into this actor branch",
            )
        entries = _read_text_items_source(
            session_root=session_dir,
            inline_items=values,
            from_file=args.from_file,
            use_stdin=args.use_stdin,
            input_name="oops entry text",
        )
        for entry in entries:
            append_oops_record(
                session_dir,
                message=_normalize_text(entry, input_name="oops entry text"),
                actor=writer,
                surface=args.surface or "",
                command=args.command or "",
                kind=args.kind or "",
                affordance=args.affordance or "",
                workaround=args.workaround or "",
                severity=args.severity or "medium",
                reproducibility=args.reproducibility,
                resolution_state=args.resolution_state,
            )
        print(f"extended oops entries: {len(entries)} item(s)")
        return 0

    session_dir, scoped_actor = session_scope._observation_scope(
        explicit_session=getattr(args, "session", None),
        explicit_actor=getattr(args, "actor", None),
    )
    explicit_actor = getattr(args, "actor", None)
    if scoped_actor:
        records = filtered_oops_records(
            oops_records(session_dir),
            surface=args.surface or "",
            command=args.command or "",
            kind=args.kind or "",
            severity=args.severity or "",
        )
        records = [
            record
            for record in records
            if writer_role(
                session_dir, scoped_actor, writer=str(record.get("actor") or "")
            )
            != "foreign"
        ]
        actor_ids = [scoped_actor]
    else:
        actor_ids, records = _aggregate_oops_records(
            session_dir,
            surface=args.surface or "",
            command=args.command or "",
            kind=args.kind or "",
            severity=args.severity or "",
        )
        if not actor_ids:
            if not _is_exact_session_root(session_dir):
                raise SystemExit(
                    "no actors bound for this session; bind one intentionally with "
                    + session_registry._actor_bind_examples(prefix="gotta actor bind")
                )
            records = filtered_oops_records(
                oops_records(session_dir),
                surface=args.surface or "",
                command=args.command or "",
                kind=args.kind or "",
                severity=args.severity or "",
            )
            payload = {
                "session_root": str(session_dir),
                "actor_count": 0,
                "actors": [],
                "shown_count": len(_limited_records(records, limit=args.limit)),
                "entries": _limited_records(records, limit=args.limit),
                "state_path": str(oops_log_path(session_dir)),
                "locator": session_charter._native_surface_locator("oops"),
                "follow_command": session_charter._native_surface_follow_command(
                    "oops"
                ),
                **oops_summary(records),
            }
            if args.output == "json":
                print(json.dumps(payload, indent=2, sort_keys=True))
                return 0
            print(f"oops: {payload['follow_command']}")
            print(
                f"entries: {payload['entry_count']} (showing {payload['shown_count']})"
            )
            print(f"severity_counts: {payload['severity_counts']}")
            print(f"kind_counts: {payload['kind_counts']}")
            print(f"surface_counts: {payload['surface_counts']}")
            print(f"resolution_counts: {payload['resolution_counts']}")
            print(f"reproducibility_counts: {payload['reproducibility_counts']}")
            print(f"affordance_counts: {payload['affordance_counts']}")
            for record in payload["entries"]:
                timestamp = str(record.get("timestamp") or "unknown-time")
                actor = str(record.get("actor") or "session")
                message = (
                    str(record.get("message") or "").strip() or "unspecified oops entry"
                )
                print(f"- `{timestamp}` [{actor}] {message}")
            return 0
    payload = {
        **_read_payload(
            session_dir,
            actor_ids,
            records,
            explicit_actor=explicit_actor,
            scoped_actor=scoped_actor,
            limit=args.limit,
        )
    }
    if args.output == "json":
        print(json.dumps(payload, indent=2, sort_keys=True))
        return 0
    if explicit_actor or scoped_actor:
        print(f"oops: {payload['follow_command']}")
    else:
        print(f"oops: session-wide across {len(actor_ids)} actor(s)")
    print(f"entries: {payload['entry_count']} (showing {payload['shown_count']})")
    print(f"severity_counts: {payload['severity_counts']}")
    print(f"kind_counts: {payload['kind_counts']}")
    print(f"surface_counts: {payload['surface_counts']}")
    print(f"resolution_counts: {payload['resolution_counts']}")
    print(f"reproducibility_counts: {payload['reproducibility_counts']}")
    print(f"affordance_counts: {payload['affordance_counts']}")
    for record in payload["entries"]:
        actor = str(record.get("actor") or "unknown-actor")
        print(
            "- "
            f"{record.get('timestamp') or 'unknown-time'} "
            f"[{actor}/{record.get('severity') or 'unknown'}] "
            f"{record.get('kind') or 'general'} "
            f"{record.get('surface') or 'unspecified'} :: "
            f"{record.get('message') or ''}"
        )
    return 0


def run(argv: list[str], *, command_name: str = "gotta oops") -> int:
    if is_long_help_request(argv):
        sys.stdout.write(format_long_help(build_parser(command_name)))
        return 0
    parser = build_parser(command_name)
    args = parser.parse_args(argv)
    return cmd_oops(args)


def main(argv: list[str] | None = None) -> int:
    return run(list(sys.argv[1:] if argv is None else argv))



"""Dispatch orchestration for packaged gotta commands."""

from __future__ import annotations

from collections.abc import Callable
import sys

from gotta.builtin import (
    PluginSpec,
    available_plugins as discovered_plugin_names,
    get_plugin,
)
from gotta.content.model import CommonOptions, ContentError, ResolvedDirs
from gotta.dispatch.budget import (
    emit_budgeted_output,
    requested_output_format,
)
from gotta.dispatch.materialize import _materialize_invocation
from gotta.dispatch.option import (
    split_common_options,
    strip_full_output_flag,
    strip_quiet_flag,
)
from gotta.dispatch.receipt import (
    _emit_receipt,
    _receipt_extra,
    _receipt_payload,
    _rerun_full_output_command,
    _result_follow_command,
    _should_emit_receipt,
)
from gotta.dispatch.runtime import (
    _captured_execution,
    _run_callable,
    _runtime_dirs,
    _sessionless_notice_bytes,
    _streams_live,
    require_operational_session,
    scoped_runtime_env,
    system_exit_status,
)
from gotta.dispatch.stream import capture_stderr, capture_stdout, _emit_captured_stderr
from gotta.resolve.intent import session_access_mode
from gotta.resolve.invoke import (
    canonical_locator,
    infer_content_type,
    invocation_locator,
    preferred_name as derive_preferred_name,
    resolve_invocation,
    should_materialize,
)

__all__ = [
    "available_plugins",
    "plugin_spec",
    "print_usage",
    "load_plugin_runner",
    "should_materialize",
    "invocation_locator",
    "canonical_locator",
    "derive_preferred_name",
    "infer_content_type",
    "session_access_mode",
    "resolve_invocation",
    "split_common_options",
    "capture_stdout",
    "capture_stderr",
    "emit_budgeted_output",
    "require_operational_session",
    "run_plugin",
    "_materialize_invocation",
]


def die(message: str, code: int = 2) -> int:
    print(message, file=sys.stderr)
    return code


def available_plugins() -> list[str]:
    return discovered_plugin_names()


def plugin_spec(plugin: str) -> PluginSpec | None:
    return get_plugin(plugin)


def print_usage() -> int:
    print("usage: gotta <plugin> [args...]", file=sys.stderr)
    print("", file=sys.stderr)
    print("canonical session-binding path: `gotta ...`", file=sys.stderr)
    print("", file=sys.stderr)
    print(
        "builtin non-session surfaces: `gotta version`, `gotta --version`, `gotta search`",
        file=sys.stderr,
    )
    print("", file=sys.stderr)
    print(
        "session investigative surfaces live under `gotta session`: "
        "`manifest`, `timeline`, `graph`, `leads`, `analyze`, `scan`",
        file=sys.stderr,
    )
    print("", file=sys.stderr)
    print("global flags: `--quiet`, `--full-output`", file=sys.stderr)
    print("", file=sys.stderr)
    print("use `gotta --help-all` for recursive command help", file=sys.stderr)
    print("", file=sys.stderr)
    print("available plugins:", file=sys.stderr)
    for plugin in available_plugins():
        spec = plugin_spec(plugin)
        description = spec.description if spec else ""
        if description:
            print(f"  - {plugin:<10} {description}", file=sys.stderr)
            continue
        print(f"  - {plugin}", file=sys.stderr)
    return 0


def load_plugin_runner(plugin: str) -> Callable[[list[str]], int]:
    spec = plugin_spec(plugin)
    if spec is None:
        raise KeyError(plugin)
    return spec.runner


def run_plugin(plugin: str, argv: list[str]) -> int:
    quiet, argv = strip_quiet_flag(argv)
    full_output, argv = strip_full_output_flag(argv)
    if plugin == "session":
        options = CommonOptions()
        cleaned = argv
    else:
        try:
            options, cleaned = split_common_options(
                argv,
                strip_actor=plugin
                in {
                    "read",
                    "confluence",
                    "gdocs",
                    "gdrive",
                    "github",
                    "grafana",
                    "granola",
                    "gsheets",
                    "jira",
                    "slack",
                },
            )
        except ContentError as exc:
            return die(str(exc))

    try:
        runner = load_plugin_runner(plugin)
    except KeyError:
        plugins = ", ".join(available_plugins())
        return die(f"unknown gotta plugin: {plugin}. available plugins: {plugins}")
    except RuntimeError as exc:
        return die(str(exc), code=1)

    try:
        resolved = resolve_invocation(plugin, cleaned, options)
    except SystemExit as exc:
        return system_exit_status(exc)
    except ContentError as exc:
        return die(str(exc))
    except RuntimeError as exc:
        return die(str(exc), code=1)
    access = session_access_mode(plugin, cleaned)
    spec = plugin_spec(plugin)
    follow_command = ""
    rerun_command = _rerun_full_output_command(plugin, cleaned)
    runtime_dirs: ResolvedDirs | None = None
    budget_output = not full_output

    def emit_success(
        stdout_data: bytes,
        *,
        stderr_data: bytes = b"",
        result: object | None = None,
        dirs: ResolvedDirs | None = None,
    ) -> int:
        nonlocal follow_command
        follow_command = _result_follow_command(result)
        emitted = emit_budgeted_output(
            stdout_data,
            output_format=requested_output_format(plugin, cleaned, stdout_data),
            budget_output=budget_output,
            follow_command=rerun_command,
        )
        if stderr_data and not quiet:
            _emit_captured_stderr(stderr_data)
        if _should_emit_receipt(plugin, cleaned):
            _emit_receipt(
                _receipt_payload(
                    emitted=emitted,
                    result=result,
                    extra=_receipt_extra(plugin, cleaned, dirs=dirs),
                ),
                quiet=quiet,
            )
        return 0

    def replay_stdout(stdout_data: bytes) -> None:
        if not stdout_data:
            return
        emit_budgeted_output(
            stdout_data,
            output_format=requested_output_format(plugin, cleaned, stdout_data),
            budget_output=budget_output,
            follow_command=rerun_command,
        )

    if not resolved.should_materialize and _streams_live(plugin, cleaned):
        if access != "none" and (
            options.session_dir or options.content_dir or options.actor
        ):
            try:
                runtime_dirs = _runtime_dirs(options, access=access)
                require_operational_session(runtime_dirs)
            except ContentError as exc:
                return die(str(exc))
            with scoped_runtime_env(runtime_dirs):
                return _run_callable(runner, cleaned)
        return _run_callable(runner, cleaned)

    if not resolved.should_materialize:
        if access != "none" and (
            options.session_dir or options.content_dir or options.actor
        ):
            try:
                runtime_dirs = _runtime_dirs(options, access=access)
                require_operational_session(runtime_dirs)
            except ContentError as exc:
                return die(str(exc))
            with scoped_runtime_env(runtime_dirs):
                with (
                    capture_stdout(preserve_tty=True) as stdout_capture,
                    capture_stderr(preserve_tty=True) as stderr_capture,
                ):
                    code = _run_callable(runner, cleaned)
        else:
            with (
                capture_stdout(preserve_tty=True) as stdout_capture,
                capture_stderr(preserve_tty=True) as stderr_capture,
            ):
                code = _run_callable(runner, cleaned)
        stderr_data = stderr_capture.getvalue()
        stdout_data = stdout_capture.getvalue()
        if code == 0:
            if not quiet:
                stderr_data += _sessionless_notice_bytes(resolved)
            return emit_success(stdout_data, stderr_data=stderr_data, dirs=runtime_dirs)
        replay_stdout(stdout_data)
        _emit_captured_stderr(stderr_data)
        return code

    try:
        runtime_dirs = _runtime_dirs(options, access=access)
        require_operational_session(runtime_dirs)
    except ContentError as exc:
        return die(str(exc))

    if (
        spec
        and spec.capture is not None
        and spec.project is not None
        and resolved.artifact_intent in {"evidence", "discovery"}
    ):
        try:
            with scoped_runtime_env(runtime_dirs):
                with capture_stderr(preserve_tty=True) as stderr_capture:
                    capture, display = _captured_execution(plugin, cleaned, options)
        except NotImplementedError:
            capture = None
            display = None
        except SystemExit as exc:
            return system_exit_status(exc)
        except (ContentError, RuntimeError) as exc:
            return die(str(exc), code=1)
        if capture is not None and display is not None:
            try:
                result = _materialize_invocation(
                    resolved,
                    capture.data,
                    options=options,
                    capture=capture,
                    dirs=runtime_dirs,
                )
            except ContentError as exc:
                return die(str(exc), code=1)
            return emit_success(
                display,
                stderr_data=stderr_capture.getvalue(),
                result=result,
                dirs=runtime_dirs,
            )

    with scoped_runtime_env(runtime_dirs):
        with (
            capture_stdout(preserve_tty=True) as stdout_capture,
            capture_stderr(preserve_tty=True) as stderr_capture,
        ):
            code = _run_callable(runner, cleaned)
    data = stdout_capture.getvalue()
    stderr_data = stderr_capture.getvalue()
    if code == 0:
        try:
            result = _materialize_invocation(
                resolved, data, options=options, dirs=runtime_dirs
            )
        except ContentError as exc:
            return die(str(exc), code=1)
        return emit_success(
            data, stderr_data=stderr_data, result=result, dirs=runtime_dirs
        )
    replay_stdout(data)
    _emit_captured_stderr(stderr_data)
    return code



"""Invocation resolution and metadata derivation."""

from __future__ import annotations

from dataclasses import dataclass

from gotta.content.model import CommonOptions
from gotta.resolve.canon import canonical_locator as derive_canonical_locator
from gotta.resolve.canon import invocation_locator as derive_invocation_locator
from gotta.resolve.intent import (
    ArtifactIntent,
    HELP_TOKENS,
    artifact_intent,
    materialization_enabled,
)
from gotta.resolve.name import infer_content_type as derive_content_type
from gotta.resolve.name import preferred_name as derive_preferred_name
from gotta.resolve.read import resolve_read_target
from gotta.resolve.search import SearchRouteError, resolve_search_route


@dataclass(frozen=True, slots=True)
class ResolvedInvocation:
    entry_plugin: str
    entry_argv: list[str]
    resolved_plugin: str
    resolved_argv: list[str]
    canonical_locator: str
    preferred_name: str
    content_type: str
    artifact_intent: ArtifactIntent
    artifact_kind: str
    should_materialize: bool
    provider: str


def _artifact_kind(intent: ArtifactIntent) -> str:
    return "" if intent == "none" else intent


def resolve_invocation(
    plugin: str,
    argv: list[str],
    options: CommonOptions | None = None,
) -> ResolvedInvocation:
    options = options or CommonOptions()
    entry_plugin = plugin
    entry_argv = list(argv)
    resolved_plugin = plugin
    resolved_argv = list(argv)
    if any(arg in HELP_TOKENS for arg in argv):
        preferred = options.save_as or f"{plugin}.txt"
        return ResolvedInvocation(
            entry_plugin=entry_plugin,
            entry_argv=entry_argv,
            resolved_plugin=plugin,
            resolved_argv=resolved_argv,
            canonical_locator=f"{plugin}:help",
            preferred_name=preferred,
            content_type=derive_content_type(plugin, argv, preferred),
            artifact_intent="none",
            artifact_kind="",
            should_materialize=False,
            provider=plugin,
        )
    if plugin == "read":
        try:
            resolved_target = resolve_read_target(argv, options)
        except SystemExit:
            target = next((item for item in argv if not item.startswith("-")), "")
            preferred = options.save_as or "read.txt"
            return ResolvedInvocation(
                entry_plugin=entry_plugin,
                entry_argv=entry_argv,
                resolved_plugin="read",
                resolved_argv=resolved_argv,
                canonical_locator=target or "read",
                preferred_name=preferred,
                content_type=derive_content_type("read", resolved_argv, preferred),
                artifact_intent="none",
                artifact_kind="",
                should_materialize=False,
                provider="read",
            )
        if resolved_target.routed_plugin is not None:
            resolved_plugin = resolved_target.routed_plugin
            resolved_argv = list(resolved_target.routed_argv)
        provider = resolved_plugin
        canonical = resolved_target.canonical_locator
        preferred = resolved_target.preferred_name
        content_type = derive_content_type(resolved_plugin, resolved_argv, preferred)
        if resolved_target.routed_plugin is not None:
            routed_intent = artifact_intent(resolved_plugin, resolved_argv)
            intent: ArtifactIntent = (
                routed_intent if resolved_target.should_materialize else "none"
            )
        else:
            intent = "evidence" if resolved_target.should_materialize else "none"
        return ResolvedInvocation(
            entry_plugin=entry_plugin,
            entry_argv=entry_argv,
            resolved_plugin=resolved_plugin,
            resolved_argv=resolved_argv,
            canonical_locator=canonical,
            preferred_name=preferred,
            content_type=content_type,
            artifact_intent=intent,
            artifact_kind=_artifact_kind(intent),
            should_materialize=intent != "none" and materialization_enabled(),
            provider=provider,
        )
    if plugin == "search":
        try:
            search_route = resolve_search_route(argv)
        except SearchRouteError:
            preferred = options.save_as or "search.md"
            return ResolvedInvocation(
                entry_plugin=entry_plugin,
                entry_argv=entry_argv,
                resolved_plugin="search",
                resolved_argv=resolved_argv,
                canonical_locator="search",
                preferred_name=preferred,
                content_type="text/markdown",
                artifact_intent="none",
                artifact_kind="",
                should_materialize=False,
                provider="search",
            )
        canonical = derive_canonical_locator(
            search_route.provider,
            search_route.provider_argv,
        )
        preferred = derive_preferred_name(
            search_route.provider,
            search_route.provider_argv,
            options,
        )
        content_type = derive_content_type(
            search_route.provider,
            search_route.provider_argv,
            preferred,
        )
        intent: ArtifactIntent = "discovery"
        return ResolvedInvocation(
            entry_plugin=entry_plugin,
            entry_argv=entry_argv,
            resolved_plugin=search_route.provider,
            resolved_argv=search_route.provider_argv,
            canonical_locator=canonical,
            preferred_name=preferred,
            content_type=content_type,
            artifact_intent=intent,
            artifact_kind=_artifact_kind(intent),
            should_materialize=materialization_enabled(),
            provider=search_route.provider,
        )
    canonical = derive_canonical_locator(plugin, argv)
    preferred = derive_preferred_name(plugin, argv, options)
    content_type = derive_content_type(plugin, argv, preferred)
    intent = artifact_intent(plugin, argv)
    return ResolvedInvocation(
        entry_plugin=entry_plugin,
        entry_argv=entry_argv,
        resolved_plugin=plugin,
        resolved_argv=argv,
        canonical_locator=canonical,
        preferred_name=preferred,
        content_type=content_type,
        artifact_intent=intent,
        artifact_kind=_artifact_kind(intent),
        should_materialize=intent != "none" and materialization_enabled(),
        provider=plugin,
    )


def should_materialize(plugin: str, argv: list[str]) -> bool:
    return resolve_invocation(plugin, argv, CommonOptions()).should_materialize


def invocation_locator(plugin: str, argv: list[str]) -> str:
    return derive_invocation_locator(plugin, argv)


def canonical_locator(plugin: str, argv: list[str]) -> str:
    return resolve_invocation(plugin, argv, CommonOptions()).canonical_locator


def preferred_name(plugin: str, argv: list[str], options: CommonOptions) -> str:
    return resolve_invocation(plugin, argv, options).preferred_name


def infer_content_type(plugin: str, argv: list[str], name: str) -> str:
    resolved = resolve_invocation(plugin, argv, CommonOptions())
    if resolved.preferred_name == name:
        return resolved.content_type
    return derive_content_type(resolved.resolved_plugin, resolved.resolved_argv, name)



"""Session activity, actor events, and note/evidence summaries."""

from __future__ import annotations

import json
import os
from pathlib import Path
import time

from gotta.compat import UTC, datetime
from gotta.content.activity import append_activity_event
from gotta.content.context import current_actor
from gotta.content.file import PRIVATE_FILE_MODE, ensure_private_dir
from gotta.content.scope import session_identity
from gotta.friction import oops_log_path
from gotta.logs import append_log_record, logs_state_path
from gotta.notes import visible_actor_notes_records
from gotta.actor import (
    writer_role,
)
from gotta.todo import (
    todo_state_path,
)

from .charter import (
    _native_surface_follow_command,
    _native_surface_locator,
    _native_surface_preferred_name,
    _session_relative_locator,
    _surface_actor_scope,
)
from .registry import (
    ACTOR_STALL_SECONDS,
    WANT_FILE,
    _actor_events_path,
    _normalize_actor_name,
    _read_actor_state,
    _write_actor_state,
)


def _append_chunk(path: Path, chunk: str) -> None:
    ensure_private_dir(path.parent)
    fd = os.open(path, os.O_APPEND | os.O_CREAT | os.O_WRONLY, PRIVATE_FILE_MODE)
    try:
        os.write(fd, chunk.encode("utf-8"))
    finally:
        os.close(fd)
    try:
        path.chmod(PRIVATE_FILE_MODE)
    except OSError:
        pass


def _append_jsonl(path: Path, payload: dict[str, object]) -> None:
    _append_chunk(path, json.dumps(payload, sort_keys=True) + "\n")


def _record_session_activity(
    work_dir: Path,
    *,
    plugin: str,
    surface: str,
    action: str,
    actor: str = "",
    target_actor: str = "",
    target: Path | None = None,
    locator: str = "",
    preferred_name: str = "",
    follow_command: str = "",
    detail: str = "",
) -> None:
    if target is not None:
        resolved = target.resolve()
        actor_scope = _surface_actor_scope(work_dir)
        if resolved == (work_dir / WANT_FILE).resolve():
            resolved_locator = _native_surface_locator("want", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "want", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "want", actor_name=actor_scope
            )
        elif resolved == (work_dir / "GOAL.md").resolve():
            resolved_locator = _native_surface_locator("goal", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "goal", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "goal", actor_name=actor_scope
            )
        elif resolved == todo_state_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("todo", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "todo", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "todo", actor_name=actor_scope
            )
        elif resolved == logs_state_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("logs", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "logs", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "logs", actor_name=actor_scope
            )
        elif resolved == oops_log_path(work_dir).resolve():
            resolved_locator = _native_surface_locator("oops", actor_name=actor_scope)
            resolved_name = _native_surface_preferred_name(
                "oops", actor_name=actor_scope
            )
            resolved_follow = _native_surface_follow_command(
                "oops", actor_name=actor_scope
            )
        else:
            resolved_locator = _session_relative_locator(work_dir, resolved)
            resolved_name = resolved.name
            resolved_follow = f"gotta read {resolved_locator!r}"
    else:
        resolved_locator = locator.strip() or f"{plugin}:{surface}"
        resolved_name = preferred_name.strip() or resolved_locator
        resolved_follow = follow_command.strip()
    activity_actor = actor.strip() or current_actor(
        default_actor=session_identity(work_dir)
    )
    payload = {
        "plugin": plugin,
        "surface": surface,
        "action": action,
        "actor": activity_actor,
        "locator": resolved_locator,
        "preferred_name": preferred_name.strip() or resolved_name,
        "follow_command": follow_command.strip() or resolved_follow,
        "detail": detail,
        "time_field": "session_recorded_at",
    }
    normalized_target = (
        _normalize_actor_name(target_actor) if target_actor.strip() else ""
    )
    if normalized_target and normalized_target != activity_actor:
        payload["target_actor"] = normalized_target
    append_activity_event(work_dir, payload)


def _append_actor_event(
    work_dir: Path,
    actor_name: str,
    *,
    event: str,
    detail: str = "",
    extra: dict[str, object] | None = None,
    author: str = "",
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    event_author = author.strip() or current_actor(default_actor=normalized_actor)
    payload: dict[str, object] = {
        "timestamp": datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
        "actor": normalized_actor,
        "author": event_author,
        "event": event,
        "detail": detail,
    }
    if extra:
        payload.update(extra)
    _append_jsonl(_actor_events_path(work_dir, actor_name), payload)
    if event != "heartbeat":
        _record_session_activity(
            work_dir,
            plugin="actor",
            surface="actor.lifecycle",
            action=event,
            actor=event_author,
            target_actor=normalized_actor,
            locator=f"actor:{payload['actor']}",
            preferred_name=str(payload["actor"]),
            follow_command=f"gotta actor status {payload['actor']}",
            detail=detail or event,
        )


def _actor_log_line(
    session_root: Path, actor_name: str, message: str, *, author: str = ""
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    log_author = author.strip() or current_actor(default_actor=normalized_actor)
    if log_author == normalized_actor:
        rendered = f"[{normalized_actor}] {message}"
    else:
        rendered = f"[{log_author} -> {normalized_actor}] {message}"
    append_log_record(session_root, message=rendered, actor=log_author)


def _record_actor_surface_activity(
    session_root: Path,
    *,
    actor_name: str,
    surface: str,
    action: str,
    detail: str,
    actor: str = "",
) -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    _record_session_activity(
        session_root,
        plugin="actor",
        surface=surface,
        action=action,
        actor=actor.strip() or current_actor(default_actor=normalized_actor),
        target_actor=normalized_actor,
        locator=_native_surface_locator(surface, actor_name=normalized_actor),
        preferred_name=_native_surface_preferred_name(
            surface, actor_name=normalized_actor
        ),
        follow_command=_native_surface_follow_command(
            surface, actor_name=normalized_actor
        ),
        detail=detail,
    )


def _actor_evidence_summary(work_dir: Path, actor_name: str) -> dict[str, object]:
    manifest_path = work_dir / "content" / "manifest.jsonl"
    entries: list[dict[str, object]] = []
    if manifest_path.exists():
        for raw_line in manifest_path.read_text(encoding="utf-8").splitlines():
            if not raw_line.strip():
                continue
            try:
                payload = json.loads(raw_line)
            except json.JSONDecodeError:
                continue
            if isinstance(payload, dict) and str(
                payload.get("actor") or session_identity(work_dir)
            ) == _normalize_actor_name(actor_name):
                entries.append(payload)
    ordered = sorted(
        entries,
        key=lambda entry: (
            str(entry.get("fetched_at") or ""),
            str(entry.get("checksum") or ""),
        ),
        reverse=True,
    )
    return {
        "artifact_count": len(entries),
        "last_artifact_at": str(ordered[0].get("fetched_at") or "").strip()
        if ordered
        else "",
        "recent_artifacts": [
            {
                "locator": str(
                    entry.get("canonical_locator") or entry.get("locator") or ""
                ).strip(),
                "preferred_name": str(entry.get("preferred_name") or "data").strip(),
                "fetched_at": str(entry.get("fetched_at") or "").strip(),
            }
            for entry in ordered[:5]
        ],
    }


def _actor_evidence_note(evidence: dict[str, object]) -> str:
    artifact_count = int(evidence.get("artifact_count") or 0)
    if artifact_count <= 0:
        return ""
    noun = "artifact" if artifact_count == 1 else "artifacts"
    return (
        f"{artifact_count} actor-attributed {noun} already live in session manifest, "
        "timeline, leads, and graph."
    )


def _actor_activity_summary(
    event: str,
    detail: str,
    *,
    author: str = "",
    target_actor: str = "",
) -> str:
    cleaned_detail = detail.strip()
    author_prefix = ""
    if author and target_actor and author != target_actor:
        author_prefix = f"{author}: "
    if event == "note":
        return (
            (author_prefix + cleaned_detail)
            if cleaned_detail
            else (author_prefix + "note").strip()
        )
    label = event.replace("_", " ")
    if cleaned_detail:
        return f"{label}: {author_prefix}{cleaned_detail}".strip()
    return f"{label}: {author_prefix}".strip(": ")


def _actor_event_records(work_dir: Path, actor_name: str) -> list[dict[str, object]]:
    path = _actor_events_path(work_dir, actor_name)
    if not path.exists():
        return []
    normalized_actor = _normalize_actor_name(actor_name)
    events: list[dict[str, object]] = []
    for index, raw_line in enumerate(path.read_text(encoding="utf-8").splitlines()):
        if not raw_line.strip():
            continue
        try:
            payload = json.loads(raw_line)
        except json.JSONDecodeError:
            continue
        if not isinstance(payload, dict):
            continue
        event = str(payload.get("event") or "").strip()
        if not event or event == "heartbeat":
            continue
        timestamp = str(payload.get("timestamp") or "").strip()
        detail = str(payload.get("detail") or "").strip()
        author = str(payload.get("author") or "").strip()
        if (
            writer_role(work_dir, normalized_actor, writer=author or normalized_actor)
            == "foreign"
        ):
            continue
        events.append(
            {
                "timestamp": timestamp,
                "event": event,
                "author": author,
                "detail": detail,
                "summary": _actor_activity_summary(
                    event,
                    detail,
                    author=author,
                    target_actor=normalized_actor,
                ),
                "_order": index,
            }
        )
    return events


def _actor_recent_activity(
    work_dir: Path, actor_name: str, *, limit: int = 5
) -> dict[str, object]:
    events = _actor_event_records(work_dir, actor_name)
    if not events:
        return {
            "recent_activity": [],
            "recent_lifecycle": [],
            "last_lifecycle_at": "",
            "last_lifecycle_summary": "",
        }
    ordered = sorted(
        events,
        key=lambda item: (
            str(item.get("timestamp") or ""),
            int(item.get("_order") or 0),
        ),
        reverse=True,
    )
    lifecycle = [item for item in ordered if str(item.get("event") or "") != "note"]
    recent_activity = [
        {
            "timestamp": str(item.get("timestamp") or ""),
            "event": str(item.get("event") or ""),
            "author": str(item.get("author") or ""),
            "detail": str(item.get("detail") or ""),
            "summary": str(item.get("summary") or ""),
        }
        for item in ordered[:limit]
    ]
    recent_lifecycle = [
        {
            "timestamp": str(item.get("timestamp") or ""),
            "event": str(item.get("event") or ""),
            "author": str(item.get("author") or ""),
            "detail": str(item.get("detail") or ""),
            "summary": str(item.get("summary") or ""),
        }
        for item in lifecycle[:limit]
    ]
    latest = recent_lifecycle[0] if recent_lifecycle else {}
    return {
        "recent_activity": recent_activity,
        "recent_lifecycle": recent_lifecycle,
        "last_lifecycle_at": str(latest.get("timestamp") or ""),
        "last_lifecycle_summary": str(latest.get("summary") or ""),
    }


def _actor_note_summary(work_dir: Path, actor_name: str) -> dict[str, object]:
    normalized_actor = _normalize_actor_name(actor_name)
    notes: list[dict[str, str]] = []
    for record in visible_actor_notes_records(work_dir, normalized_actor):
        if str(record.get("author") or "").strip() != normalized_actor:
            continue
        timestamp = str(record.get("timestamp") or "").strip()
        message = str(record.get("message") or "").strip()
        if not timestamp or not message:
            continue
        first_line = message.splitlines()[0] if message.splitlines() else message
        notes.append(
            {
                "timestamp": timestamp,
                "summary": first_line.strip() or message,
            }
        )
    ordered = sorted(
        notes, key=lambda item: str(item.get("timestamp") or ""), reverse=True
    )
    latest = ordered[0] if ordered else {}
    notes_stale = False
    latest_timestamp = str(latest.get("timestamp") or "")
    if latest_timestamp:
        try:
            latest_dt = datetime.fromisoformat(latest_timestamp.replace("Z", "+00:00"))
        except ValueError:
            latest_dt = None
        if latest_dt is not None:
            notes_stale = (time.time() - latest_dt.timestamp()) > ACTOR_STALL_SECONDS
    return {
        "last_note_at": latest_timestamp,
        "last_note_summary": str(latest.get("summary") or ""),
        "notes_stale": notes_stale,
    }


def _actor_note_check_summary(work_dir: Path, actor_name: str) -> dict[str, object]:
    state = _read_actor_state(work_dir, actor_name)
    try:
        count = int(state.get("note_checks_since_update") or 0)
    except (TypeError, ValueError):
        count = 0
    return {
        "note_checks_since_update": max(count, 0),
        "last_note_check_at": str(state.get("last_note_check_at") or ""),
        "last_note_check_by": str(state.get("last_note_check_by") or ""),
    }


def _reset_note_check_feedback(work_dir: Path, actor_name: str) -> None:
    _write_actor_state(
        work_dir,
        actor_name,
        {
            "note_checks_since_update": 0,
            "last_note_check_at": None,
            "last_note_check_by": None,
        },
    )


def _record_note_check(work_dir: Path, actor_name: str, *, reader: str = "") -> None:
    normalized_actor = _normalize_actor_name(actor_name)
    normalized_reader = _normalize_actor_name(
        reader.strip() or current_actor(default_actor="")
    )
    if not normalized_reader or normalized_reader == normalized_actor:
        return
    summary = _actor_note_check_summary(work_dir, normalized_actor)
    _write_actor_state(
        work_dir,
        normalized_actor,
        {
            "note_checks_since_update": int(
                summary.get("note_checks_since_update") or 0
            )
            + 1,
            "last_note_check_at": datetime.now(tz=UTC).strftime("%Y-%m-%dT%H:%M:%SZ"),
            "last_note_check_by": normalized_reader,
        },
    )



#!/usr/bin/env python3
"""Canonical gotta CLI entrypoint."""

from __future__ import annotations

import os
import sys

from gotta.actors import resolve_actor_context, seed_actor_context
from gotta.actor import SESSION_ACTOR_ENV
import gotta.cli.argv as cli_argv
import gotta.cli.bind as cli_bind
import gotta.cli.env as cli_env
import gotta.cli.notice as cli_notice
import gotta.cli.select as cli_select
from gotta.dispatch.materialize import SUPPRESS_MATERIALIZATION_ENV
from gotta import binding as binding_helpers
from gotta.content.context import current_context_binding
from gotta.content.scope import (
    resolve_session_reference,
    session_identity,
    session_is_initialized,
)
from gotta import topology
from gotta.session import registry as session_registry


def main(argv: list[str] | None = None) -> int:
    args = list(sys.argv[1:] if argv is None else argv)
    normalized = cli_argv.normalize_help_aliases(args)
    effective = cli_argv._argv_without_global_flags(normalized)
    try:
        if cli_argv._is_nonbinding_help(effective) or cli_argv._is_version_request(
            effective
        ):
            return cli_argv._gotta_main(normalized)
        plugin_name = effective[0] if effective else ""
        context = current_context_binding()
        context_id = context.context_id
        context_source = context.context_source
        explicit_session = cli_argv._explicit_session_arg(effective)
        explicit_actor = cli_argv._explicit_actor_arg(effective)
        explicit_target = bool(explicit_session or explicit_actor)
        session_access = cli_argv._session_access_mode(effective)
        auto_bootstrap = cli_select._should_auto_bootstrap_session(
            argv=effective,
            context_source=context_source,
            explicit_session=explicit_session,
            session_access=session_access,
        )
        shared_root_command = cli_select._uses_shared_session_root(
            argv=effective,
            explicit_session=explicit_session,
            explicit_actor=explicit_actor,
            session_access=session_access,
        )
        created = False
        bound_current_context = False
        init_command = (
            plugin_name == "session" and len(effective) >= 2 and effective[1] == "init"
        )
        if (
            plugin_name == "session"
            and len(effective) >= 2
            and effective[1] in {"bind"}
        ):
            return cli_argv._gotta_main(normalized)
        if session_access == "none":
            return cli_argv._gotta_main(normalized)
        if explicit_session:
            if init_command:
                target_identity = topology.normalize_identity(
                    cli_select._active_identity(context_id)
                )
                root = resolve_session_reference(
                    explicit_session,
                    identity=target_identity,
                    allow_missing=True,
                )
            elif shared_root_command:
                root = cli_select._resolve_shared_explicit_session(explicit_session)
            elif (
                session_access == "read"
                and not explicit_actor
                and cli_select._prefers_primary_actor_root(effective)
            ):
                shared_root = cli_select._resolve_shared_explicit_session(
                    explicit_session
                )
                if shared_root is not None and cli_select._is_shared_session_root(
                    shared_root
                ):
                    primary_root = cli_select._resolve_primary_actor_root(shared_root)
                    if primary_root is None:
                        return cli_notice.die(
                            "this shared session does not resolve to one canonical actor root; "
                            "pass `--actor <actor>` explicitly"
                        )
                    root = primary_root
                else:
                    root = shared_root
            else:
                target_identity = topology.normalize_identity(
                    explicit_actor or cli_select._active_identity(context_id)
                )
                explicit_root = resolve_session_reference(
                    explicit_session,
                    identity=target_identity,
                    allow_missing=False,
                )
                if explicit_root is not None and explicit_actor:
                    resolved_actor = session_registry._resolve_bound_actor_name(
                        explicit_root,
                        explicit_actor,
                    )
                    root = session_registry._actor_session_dir(
                        explicit_root, resolved_actor
                    )
                elif explicit_root is not None:
                    root = explicit_root
                else:
                    if explicit_actor:
                        return cli_notice.die(
                            "explicit actor targeting requires an existing shared session and a bound actor"
                        )
                    root = resolve_session_reference(
                        explicit_session,
                        identity=target_identity,
                        allow_missing=True,
                    )
            if root is None:
                return cli_notice.die(
                    "session references must be an absolute path, a shared session id, "
                    "or an explicit <session>/<actor> session reference"
                )
        elif explicit_actor:
            current = cli_select._prefer_bound_session_root()
            if current is None:
                current = cli_bind._resolve_existing_session_root(context_id)
            if current is None:
                return cli_notice.die(
                    "explicit actor targeting requires an existing session; run "
                    "`gotta session bind` first or pass `--session <session-id>`"
                )
            resolved_actor = session_registry._resolve_bound_actor_name(
                current,
                explicit_actor,
            )
            root = session_registry._actor_session_dir(current, resolved_actor)
        else:
            root = cli_select._prefer_bound_session_root()
        if root is None:
            root = cli_bind._resolve_existing_session_root(context_id)
        if root is None and (session_access == "write" or auto_bootstrap):
            root, created = cli_bind._bind_session_root(context_id, context_source)
            bound_current_context = True
        if root is None and session_access != "ambient":
            return cli_notice.die(
                "this command requires an existing session; run `gotta session bind` "
                "first or pass `--session <session-id>`"
            )
        requested_root = root
        if (
            session_access == "ambient"
            and root is not None
            and not session_is_initialized(root)
            and not shared_root_command
        ):
            if explicit_target:
                return cli_notice.die(
                    "ambient retrieval requires an existing initialized actor root in the "
                    "target session; bind an actor there first or pass `--actor <actor>`"
                )
            root = None
        scaffold_created = False
        if (
            auto_bootstrap
            and root is not None
            and session_access in {"read", "ambient"}
        ):
            root, scaffold_created = cli_bind._ensure_scaffolded_session(
                root,
                context_id=context_id,
                context_source=context_source,
            )
        if (
            session_access == "read"
            and root is not None
            and not session_is_initialized(root)
        ):
            if not shared_root_command:
                existing = cli_select._existing_actor_root_for_session(
                    root,
                    preferred_identities=cli_select._preferred_read_only_session_identities(
                        root,
                        explicit_actor=explicit_actor,
                    ),
                )
                if existing is not None:
                    root = existing
        if session_access == "read":
            if root is None:
                return cli_notice.die(
                    "explicit session inspection requires an initialized actor root in the "
                    "target shared session; bind an actor there first or pass --actor"
                )
            if not session_is_initialized(root) and not shared_root_command:
                if explicit_session and not cli_select._is_shared_session_root(root):
                    return cli_notice.die(
                        "explicit session inspection requires an existing initialized session "
                        "at that exact root"
                    )
                return cli_notice.die(
                    "explicit session inspection requires an initialized actor root in the "
                    "target shared session; bind an actor there first or pass --actor"
                )
        elif session_access == "ambient":
            scaffold_created = False
        else:
            assert root is not None
            root, scaffold_created = cli_bind._ensure_scaffolded_session(
                root,
                context_id=context_id,
                context_source=context_source,
            )
            if init_command and root is not None and not bound_current_context:
                root = binding_helpers.bind_root_for_context(
                    root,
                    context_id=context_id,
                    context_source=context_source,
                )
                bound_current_context = True
        created = created or scaffold_created
        original_env = os.environ.copy()
        try:
            acting_actor = resolve_actor_context(
                default_speaker=cli_select._active_identity(context_id)
            ).speaker or cli_select._active_identity(context_id)
            if root is not None:
                if shared_root_command and cli_select._is_shared_session_root(root):
                    cli_env._hydrate_shared_session_environment(
                        root,
                        context_id=context_id,
                        context_source=context_source,
                    )
                else:
                    cli_env._hydrate_environment(
                        root, context_id=context_id, context_source=context_source
                    )
                    seed_actor_context(acting_actor)
                    if explicit_actor:
                        os.environ[SESSION_ACTOR_ENV] = session_identity(root)
                if created:
                    print(
                        "\n".join(
                            cli_notice._creation_receipt_lines(
                                root,
                                context_id=context_id,
                                context_source=context_source,
                                bound=bound_current_context,
                            )
                        ),
                        file=sys.stderr,
                    )
                warning = cli_notice._actor_stop_warning(root)
                if warning:
                    print(warning, file=sys.stderr)
                elif cli_notice._should_emit_actor_note_check_warning(
                    argv=effective,
                    root=root,
                    requested_root=requested_root,
                    acting_actor=acting_actor,
                    explicit_actor=explicit_actor,
                ):
                    pulse_warning = cli_notice._actor_note_check_warning(root)
                    if pulse_warning:
                        print(pulse_warning, file=sys.stderr)
            elif session_access == "ambient":
                os.environ[SUPPRESS_MATERIALIZATION_ENV] = "1"
                os.environ[cli_env._AMBIENT_SESSIONLESS_ENV] = "1"
            return cli_argv._gotta_main(normalized)
        finally:
            os.environ.clear()
            os.environ.update(original_env)
    except BrokenPipeError:
        cli_notice._silence_stdout()
        return 0



from __future__ import annotations

import os
from pathlib import Path

from gotta.content.context import current_actor, current_context_binding
from gotta.content.env import (
    ACTOR_ID_ENV,
    CONTENT_ENV,
    SESSION_ACTOR_ENV,
    SESSION_ENV,
    SESSION_ID_ENV,
    SESSION_INITIALIZED_ENV,
    discover_state_env,
    load_state_env_at_root,
)
from gotta.content.file import ensure_private_dir
from gotta.content.model import CommonOptions, ContentError, ResolvedDirs
from gotta import topology

DEFAULT_SESSION_ROOT = topology.DEFAULT_SESSIONS_ROOT


def session_is_initialized(root: Path) -> bool:
    if topology.parse_shared_session_root(root) is not None:
        return False
    from gotta.content.env import state_env_path

    return state_env_path(root).exists()


def session_id(root: Path) -> str:
    state = load_state_env_at_root(root)
    explicit = str(state.get(SESSION_ID_ENV) or "").strip()
    if explicit:
        return explicit
    return topology.shared_session_id(root)


def resolve_session_root_by_id(session_ref: str) -> Path | None:
    root = topology.resolve_session_root_by_id(session_ref)
    if root is None:
        return None
    if session_is_initialized(root):
        return root.resolve()
    return root.resolve()


def resolve_session_reference(
    raw: str,
    *,
    identity: str | None = None,
    allow_missing: bool = False,
) -> Path | None:
    normalized = raw.strip()
    if not normalized:
        return None
    candidate = Path(normalized).expanduser()
    if candidate.is_absolute():
        resolved = candidate.resolve()
        shared_id = topology.parse_shared_session_root(resolved)
        if shared_id is not None and identity:
            return topology.session_root_for(shared_id, identity).resolve()
        return resolved
    by_id = resolve_session_root_by_id(normalized)
    if by_id is not None:
        shared_id = topology.parse_shared_session_root(by_id)
        if shared_id is not None and identity:
            return topology.session_root_for(shared_id, identity).resolve()
        return by_id
    if "/" in normalized:
        raw_session_id, session_identity = normalized.split("/", 1)
        root = topology.session_root_for(raw_session_id, session_identity)
        if allow_missing or root.exists() or root.is_symlink():
            return root.resolve()
    if identity:
        shared_root = topology.shared_session_root_for(normalized)
        if allow_missing or shared_root.exists() or shared_root.is_symlink():
            return topology.session_root_for(normalized, identity).resolve()
    return None


def context_bound_session_root() -> Path | None:
    binding = current_context_binding()
    root = topology.resolve_binding(binding.binding_id)
    if root is None:
        legacy = (
            DEFAULT_SESSION_ROOT.expanduser().resolve()
            / binding.binding_id
            / "actors"
            / binding.binding_id
        )
        if session_is_initialized(legacy):
            return legacy
        return None
    if session_is_initialized(root):
        return root
    return None


def bound_session_root(*, include_context_session: bool = True) -> Path | None:
    discovered = discover_state_env(include_context_session=include_context_session)
    identity_raw = (
        os.environ.get(ACTOR_ID_ENV, "").strip()
        or os.environ.get(SESSION_ACTOR_ENV, "").strip()
        or str(discovered.get(SESSION_ACTOR_ENV) or "").strip()
    )
    identity = (
        topology.normalize_identity(identity_raw)
        if identity_raw and not topology.is_placeholder_identity(identity_raw)
        else ""
    )
    explicit = os.environ.get(SESSION_ENV, "").strip()
    if explicit:
        candidate = resolve_session_reference(
            explicit, identity=identity, allow_missing=False
        )
        if candidate is not None and session_is_initialized(candidate):
            return candidate
    discovered_root = str(discovered.get(SESSION_ENV) or "").strip()
    if discovered_root:
        candidate = resolve_session_reference(
            discovered_root,
            identity=identity,
            allow_missing=False,
        )
        if candidate is not None and session_is_initialized(candidate):
            return candidate
    for raw_id in (
        os.environ.get(SESSION_ID_ENV, "").strip(),
        str(discovered.get(SESSION_ID_ENV) or "").strip(),
    ):
        if not raw_id:
            continue
        if identity:
            sibling = topology.session_root_for(raw_id, identity)
            if session_is_initialized(sibling):
                return sibling
        candidate = resolve_session_root_by_id(raw_id)
        if candidate is not None and session_is_initialized(candidate):
            return candidate
    if include_context_session:
        return context_bound_session_root()
    return None


def resolve_dirs(options: CommonOptions, *, create: bool) -> ResolvedDirs:
    discovered = discover_state_env()
    explicit_session = bool(options.session_dir)
    session_raw = (
        options.session_dir
        or os.environ.get(SESSION_ENV, "").strip()
        or discovered.get(SESSION_ENV, "").strip()
    )
    content_raw = (
        options.content_dir
        or ("" if explicit_session else os.environ.get(CONTENT_ENV, "").strip())
        or ("" if explicit_session else discovered.get(CONTENT_ENV, "").strip())
    )
    session_id_raw = (
        options.session_id
        or os.environ.get(SESSION_ID_ENV, "").strip()
        or discovered.get(SESSION_ID_ENV, "").strip()
    )
    identity_raw = (
        options.actor
        or os.environ.get(ACTOR_ID_ENV, "").strip()
        or os.environ.get(SESSION_ACTOR_ENV, "").strip()
        or discovered.get(SESSION_ACTOR_ENV, "").strip()
    )
    if not identity_raw or topology.is_placeholder_identity(identity_raw):
        identity_raw = current_context_binding().binding_id

    session: Path | None = (
        resolve_session_reference(
            session_raw,
            identity=identity_raw,
            allow_missing=create,
        )
        if session_raw
        else None
    )
    if session_raw and session is None:
        normalized_session_id = topology.normalize_session_id(session_raw)
        identity = topology.normalize_identity(identity_raw)
        session = topology.session_root_for(normalized_session_id, identity)
        if create:
            ensure_private_dir(shared_session_root(normalized_session_id))
        content_raw = str(shared_session_root(normalized_session_id) / "content")
    if session is None and session_id_raw:
        normalized_session_id = topology.normalize_session_id(session_id_raw)
        identity = topology.normalize_identity(identity_raw)
        session = topology.session_root_for(normalized_session_id, identity)
        if create:
            ensure_private_dir(shared_session_root(normalized_session_id))
        content_raw = str(shared_session_root(normalized_session_id) / "content")
    if not content_raw and session is not None:
        parsed = topology.parse_grouped_session_root(session)
        if parsed is not None:
            content_raw = str(shared_session_root(parsed[0]) / "content")
    content = Path(content_raw).expanduser() if content_raw else None

    if session is None and content is not None:
        session = content.parent
    if session is not None:
        content = content or (session / "content")

    if session is None or content is None:
        raise ContentError(
            "missing shared content context; gotta needs a session root and content root. "
            "Set GOTTA_SESSION_DIR / GOTTA_SESSION_CONTENT_DIR, pass --session/--content-dir, "
            "or use `gotta ...` so gotta can bind or scaffold the correct session "
            'for you. Use `gotta session init "$WS"` only when you intentionally '
            "want to scaffold one exact root."
        )

    if create:
        session = ensure_private_dir(session.resolve())
        content = ensure_private_dir(content.resolve())
    else:
        session = session.resolve()
        content = content.resolve()

    return ResolvedDirs(session, content)


def shared_session_root(session_id: str) -> Path:
    return topology.shared_session_root_for(session_id)


def session_shared_id(root: Path) -> str:
    state = load_state_env_at_root(root)
    explicit = str(state.get(SESSION_ID_ENV) or "").strip()
    if explicit:
        return topology.normalize_session_id(explicit)
    return topology.shared_session_id(root)


def session_actor_scope(root: Path) -> str:
    derived = topology.session_identity(root)
    if derived and not topology.is_placeholder_identity(derived):
        return topology.normalize_identity(derived)
    return ""


def session_identity(root: Path) -> str:
    state = load_state_env_at_root(root)
    explicit = str(
        state.get(SESSION_ACTOR_ENV) or state.get("GOTTA_SESSION_ACTOR") or ""
    ).strip()
    if explicit and not topology.is_placeholder_identity(explicit):
        return topology.normalize_identity(explicit)
    derived = topology.session_identity(root)
    if derived and not topology.is_placeholder_identity(derived):
        return derived
    resolved = root.expanduser().resolve()
    if topology.parse_shared_session_root(resolved) is not None:
        return current_actor()
    fallback = topology.normalize_identity(resolved.name)
    if topology.is_placeholder_identity(fallback):
        return current_context_binding().binding_id
    return fallback


def session_surface_initialized(root: Path) -> bool:
    state = load_state_env_at_root(root)
    if state.get(SESSION_INITIALIZED_ENV, "").strip() == "1":
        return True
    required = ("WANT.md", "GOAL.md")
    return all((root / name).exists() for name in required)



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any

from gotta.content.file import (
    append_jsonl_line,
    ensure_name_link,
    ensure_private_dir,
    read_json_object,
    timestamp_log_path,
    write_bytes_atomic,
    write_text_atomic,
)
from gotta.content.model import (
    ContentEvent,
    ContentSnapshot,
    Materialization,
    ResolvedDirs,
)
from gotta.content.path import is_sha256_digest, sanitize_name
from gotta.content.scope import session_identity
from gotta.content.stamp import iso_utc


def _append_manifest(dirs: ResolvedDirs, entry: dict[str, Any]) -> None:
    append_jsonl_line(dirs.content_dir / "manifest.jsonl", entry)


def _merge_meta(
    existing: dict[str, Any] | None, update: dict[str, Any]
) -> dict[str, Any]:
    payload: dict[str, Any] = {}
    if existing:
        payload.update(existing)
    payload.update(update)
    if existing and "created_at" in existing:
        payload["created_at"] = existing["created_at"]
    return payload


def _read_existing_meta(path: Path) -> dict[str, Any] | None:
    if not path.exists():
        return None
    return read_json_object(path)


def materialize_bytes(
    data: bytes,
    *,
    dirs: ResolvedDirs,
    preferred_name: str,
    metadata: dict[str, Any],
    timestamp: str | None = None,
) -> Materialization:
    import hashlib

    digest = hashlib.sha256(data).hexdigest()
    content_dir = dirs.content_dir / digest
    data_path = content_dir / "data"
    meta_path = content_dir / "meta.json"
    names_dir = content_dir / "names"
    logs_dir = content_dir / "logs"
    ensure_private_dir(names_dir)
    ensure_private_dir(logs_dir)

    if not data_path.exists():
        write_bytes_atomic(data_path, data)

    name_link = ensure_name_link(names_dir, data_path, preferred_name)
    ts = timestamp or iso_utc()
    fetch_link = timestamp_log_path(logs_dir, ts)
    fetch_link.symlink_to(os.path.relpath(name_link, start=logs_dir))

    existing = _read_existing_meta(meta_path)
    update = _merge_meta(
        existing,
        {
            "hash": digest,
            "algorithm": "sha256",
            "bytes": len(data),
            "preferred_name": sanitize_name(preferred_name),
            "original_name": preferred_name,
            **metadata,
        },
    )
    if not update.get("created_at"):
        update["created_at"] = ts
    update["fetched_at"] = ts
    write_text_atomic(
        meta_path,
        json.dumps(update, indent=2, sort_keys=True) + "\n",
    )
    _append_manifest(
        dirs,
        {
            "locator": metadata.get("locator", ""),
            "canonical_locator": metadata.get("canonical_locator", ""),
            "canonical_path": str(data_path),
            "fetched_at": ts,
            "fetcher": metadata.get("tool", "gotta"),
            "plugin": metadata.get("plugin", ""),
            "provider": metadata.get("provider", ""),
            "artifact_kind": metadata.get("artifact_kind", ""),
            "actor": metadata.get("actor") or session_identity(dirs.session_dir),
            "actor_dir": metadata.get("actor_dir", ""),
            "session_root": metadata.get("session_root", ""),
            "visibility_level": metadata.get("visibility_level", ""),
            "visibility_boundary": metadata.get("visibility_boundary", ""),
            "visibility_confidence": metadata.get("visibility_confidence", ""),
            "visibility_basis": metadata.get("visibility_basis", []),
            "checksum": digest,
            "preferred_name": sanitize_name(preferred_name),
            "fetch_link": str(fetch_link),
        },
    )
    import gotta.lead.cache as lead_cache

    lead_cache.maybe_write_lead_cache(content_dir)

    return Materialization(
        content_dir=content_dir,
        data_path=data_path,
        meta_path=meta_path,
        names_dir=names_dir,
        logs_dir=logs_dir,
        name_link=name_link,
        fetch_link=fetch_link,
        digest=digest,
        artifact_kind=str(metadata.get("artifact_kind", "") or "").strip(),
    )


def _collect_directory_snapshot(content_dir: Path) -> ContentSnapshot | None:
    digest = content_dir.name
    if not content_dir.is_dir() or not is_sha256_digest(digest):
        return None
    data_path = content_dir / "data"
    if not data_path.exists():
        return None
    meta_path = content_dir / "meta.json"
    names_dir = content_dir / "names"
    logs_dir = content_dir / "logs"
    names = (
        sorted(path.name for path in names_dir.iterdir() if path.is_symlink())
        if names_dir.exists()
        else []
    )
    events: list[ContentEvent] = []
    if logs_dir.exists():
        for path in sorted(logs_dir.iterdir(), key=lambda item: item.name):
            if not path.is_symlink():
                continue
            target = path.parent / os.readlink(path)
            events.append(
                ContentEvent(
                    timestamp=path.name.split("--", 1)[0],
                    link_name=target.name,
                    link_path=target.resolve(),
                    log_path=path,
                )
            )
    return ContentSnapshot(
        digest=digest,
        content_dir=content_dir,
        data_path=data_path,
        meta_path=meta_path if meta_path.exists() else None,
        names_dir=names_dir,
        logs_dir=logs_dir,
        names=names,
        events=events,
        metadata=read_json_object(meta_path if meta_path.exists() else None),
    )


def scan_content_store(content_dir: Path) -> list[ContentSnapshot]:
    snapshots: dict[str, ContentSnapshot] = {}
    if not content_dir.exists():
        return []
    for path in sorted(content_dir.iterdir(), key=lambda item: item.name):
        snapshot = _collect_directory_snapshot(path)
        if snapshot is not None:
            snapshots[snapshot.digest] = snapshot
    return [snapshots[digest] for digest in sorted(snapshots)]



"""Parsing and scope resolution for `gotta session`."""

from __future__ import annotations

import argparse

from gotta.content.env import state_env_path
from gotta.content.model import CommonOptions, ContentError, ResolvedDirs
from gotta.content.scope import resolve_dirs, session_is_initialized
from gotta.session import charter as session_charter
from gotta.session import registry as session_registry
from gotta.session import scope as session_scope

from .core import TIMELINE_MODE_HELP, shared_session_dirs_from_ref


def session_access_mode(argv: list[str]) -> str:
    positionals = session_charter.argv_positionals(
        argv,
        valued_flags=(
            "--session",
            "--actor",
            "--content-dir",
            "--output",
            "--limit",
            "--offset",
        ),
    )
    subcommand = positionals[0] if positionals else "show"
    return "write" if subcommand in {"init", "bind"} else "read"


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gotta session",
        description="Inspect the active session-rooted gotta context.",
    )
    sub = parser.add_subparsers(dest="command")

    init = sub.add_parser("init")
    bind = sub.add_parser("bind")
    show = sub.add_parser("show")
    doctor = sub.add_parser("doctor")
    manifest = sub.add_parser("manifest")
    timeline = sub.add_parser(
        "timeline",
        description=(
            "Inspect session chronology. Default mode is acquired order; "
            "explicit source-authored chronology is available through "
            "`--mode created`, `--mode updated`, or `--mode best-effort`."
        ),
    )
    graph = sub.add_parser("graph")
    analyze = sub.add_parser("analyze")
    scan = sub.add_parser(
        "scan",
        description=(
            "Search materialized text across the session corpus without "
            "dropping to shell-native grep tools."
        ),
    )
    leads = sub.add_parser(
        "leads",
        description=(
            "Inspect explicit, followable leads mined from already-materialized "
            "artifact content. Leads are shown best-first using observed "
            "signals such as native-ness, materialization, recurrence, and "
            "source context."
        ),
    )

    for parser_ in (
        init,
        show,
        doctor,
        manifest,
        timeline,
        graph,
        analyze,
        scan,
        leads,
    ):
        parser_.add_argument("--session", help="session root")
        parser_.add_argument("--actor", help="actor within the current session")
        parser_.add_argument(
            "--content-dir", help="explicit content directory override"
        )

    bind.add_argument(
        "session_id",
        nargs="?",
        help="session reference: shared session id, exact session root, or <session>/<actor>",
    )
    bind.add_argument(
        "--output", choices=["summary", "json", "path"], default="summary"
    )

    init.add_argument(
        "--output",
        "--print",
        dest="print_format",
        choices=["env", "json", "path", "sh"],
        default="path",
    )
    show.add_argument(
        "--output",
        "--print",
        dest="print_format",
        choices=["env", "json", "path", "sh"],
        default="path",
    )
    doctor.add_argument(
        "--output",
        "--print",
        dest="print_format",
        choices=["summary", "json"],
        default="json",
    )
    manifest.add_argument("--plugin")
    manifest.add_argument("--locator")
    manifest.add_argument(
        "--filter",
        help="case-insensitive regex applied to manifest rows",
    )
    manifest.add_argument("--limit", type=int, default=100)
    manifest.add_argument("--offset", type=int, default=0)
    manifest.add_argument("--all", action="store_true")
    manifest.add_argument("--output", choices=["json", "text"], default="text")
    timeline.add_argument(
        "--filter",
        help="case-insensitive regex applied to timeline events",
    )
    timeline.add_argument("--limit", type=int, default=100)
    timeline.add_argument("--offset", type=int, default=0)
    timeline.add_argument("--all", action="store_true")
    timeline.add_argument("--output", choices=["json", "text"], default="text")
    timeline.add_argument(
        "--mode",
        default="acquired",
        metavar="MODE",
        help=TIMELINE_MODE_HELP,
    )
    graph.add_argument(
        "--output",
        choices=["mermaid", "json", "text"],
        default="mermaid",
        help="render the content graph as Mermaid, text, or structured JSON",
    )
    graph.add_argument(
        "--filter",
        help="case-insensitive regex applied to graph nodes and edges",
    )
    analyze.add_argument(
        "--output",
        choices=["text", "mermaid", "markdown", "json"],
        default="text",
        help="render a text summary, Mermaid graph, Markdown bundle, or structured JSON to stdout",
    )
    analyze.add_argument(
        "--mode",
        choices=["lineage", "semantic", "all"],
        default="all",
        help="write lineage content, semantic content, or both; raw Mermaid output requires lineage or semantic mode",
    )
    analyze.add_argument(
        "--focus",
        help="narrow the analysis view to one keyword, locator, artifact name, or checksum prefix",
    )
    analyze.add_argument(
        "--limit",
        type=int,
        default=8,
        help="maximum items to show per overview/focus section",
    )
    scan.add_argument(
        "query",
        help="literal text or regex to search for in stored materialized text",
    )
    scan.add_argument("--plugin")
    scan.add_argument("--locator")
    scan.add_argument("--kind", choices=["discovery", "evidence"])
    scan.add_argument("--match", choices=["literal", "regex"], default="literal")
    scan.add_argument("--case-sensitive", action="store_true")
    scan.add_argument("--context", type=int, default=2)
    scan.add_argument("--snippets", type=int, default=3)
    scan.add_argument("--limit", type=int, default=20)
    scan.add_argument("--offset", type=int, default=0)
    scan.add_argument("--all", action="store_true")
    scan.add_argument("--output", choices=["json", "text"], default="text")
    leads.add_argument("target", nargs="?")
    leads.add_argument(
        "--filter",
        help="case-insensitive regex applied to derived lead records",
    )
    leads.add_argument("--limit", type=int, default=100)
    leads.add_argument("--offset", type=int, default=0)
    leads.add_argument("--all", action="store_true")
    leads.add_argument("--output", choices=["json", "text"], default="text")
    return parser


def options_from_args(args: argparse.Namespace) -> CommonOptions:
    return CommonOptions(
        session_dir=getattr(args, "session", None),
        content_dir=getattr(args, "content_dir", None),
        actor=getattr(args, "actor", None),
    )


def explicit_session_ref(args: argparse.Namespace) -> str:
    return str(getattr(args, "session", None) or "").strip()


def explicit_actor_ref(args: argparse.Namespace) -> str:
    return str(getattr(args, "actor", None) or "").strip()


def session_dirs_for_read(args: argparse.Namespace) -> ResolvedDirs:
    session_ref = explicit_session_ref(args)
    if session_ref:
        shared = shared_session_dirs_from_ref(session_ref)
        if shared is not None:
            shared_root, _session_id = shared
            group_root = session_registry._group_session_root(shared_root)
            if group_root != shared_root or (group_root / "actors").is_dir():
                return ResolvedDirs(
                    session_dir=group_root.resolve(),
                    content_dir=(group_root / "content").resolve(),
                )
        return resolve_dirs(
            CommonOptions(
                session_dir=session_ref,
                content_dir=getattr(args, "content_dir", None),
            ),
            create=False,
        )
    return resolve_dirs(
        CommonOptions(
            session_dir=getattr(args, "session", None),
            content_dir=getattr(args, "content_dir", None),
        ),
        create=False,
    )


def session_scope_started(dirs: ResolvedDirs) -> bool:
    if session_is_initialized(dirs.session_dir):
        return True
    session_root = session_registry._group_session_root(dirs.session_dir)
    if session_root != dirs.session_dir.resolve() and session_is_initialized(
        session_root
    ):
        return True
    if (session_root / "actors").is_dir() and dirs.content_dir.exists():
        return bool(session_scope._selected_actor_ids(session_root))
    return False


def require_started_session(dirs: ResolvedDirs) -> None:
    state_file = state_env_path(dirs.session_dir)
    if not session_scope_started(dirs):
        raise ContentError(
            "start or bind a session first with `gotta ...`. Stable interactive "
            "contexts adopt and scaffold their deterministic session on first "
            "session-aware use. Use `gotta session init` only when you "
            "intentionally want to scaffold one exact root; "
            f"missing {state_file}"
        )



#!/usr/bin/env python3
"""Durable provider configuration and readiness guidance."""

from __future__ import annotations

import argparse
from dataclasses import dataclass
import json
import sys
from typing import Any, Callable

from gotta.config import display_path, primary_config_file
from gotta.helptext import is_long_help_request, print_long_help
from gotta.providers import slack as slack_provider


class ToolError(RuntimeError):
    """Raised when the config surface cannot satisfy a request."""


@dataclass(frozen=True)
class ConfigSurface:
    name: str
    description: str
    resolve_target: Callable[[str], str]
    inspect: Callable[[str], dict[str, Any]]
    persist: Callable[[str], None]
    auto_configure: Callable[[str], None]


def die(message: str, code: int = 2) -> int:
    print(message, file=sys.stderr)
    return code


def print_json(data: Any) -> None:
    json.dump(data, sys.stdout, indent=2, sort_keys=True)
    sys.stdout.write("\n")


def is_interactive() -> bool:
    return sys.stdin.isatty() and sys.stderr.isatty()


def _slack_resolve_target(raw: str) -> str:
    workspace = slack_provider.workspace_from_target(raw)
    if workspace:
        return workspace
    raise ToolError(
        "unable to derive a Slack workspace from that target; pass a workspace slug "
        "like `demo`, a workspace-hosted Slack permalink/doc URL, or a generic Slack "
        "link when exactly one local workspace is unambiguous"
    )


def _slack_auto_configure(workspace: str) -> None:
    slack_provider.ensure_workspace_auth(workspace, interactive_ok=True)
    auth_state, _path = slack_provider.ensure_live_search_auth(
        workspace,
        interactive_ok=True,
    )
    slack_provider.slack_auth_test(workspace, auth_state)


SLACK_SURFACE = ConfigSurface(
    name="slack",
    description=(
        "persist one default Slack workspace, derive it from a link when needed, "
        "and inspect or complete Slack readiness"
    ),
    resolve_target=_slack_resolve_target,
    inspect=slack_provider.slack_status_payload,
    persist=slack_provider.persist_selected_workspace,
    auto_configure=_slack_auto_configure,
)

SURFACES = {SLACK_SURFACE.name: SLACK_SURFACE}


def _render_summary_text(payload: dict[str, Any]) -> str:
    lines = [
        f"config_file\t{display_path(primary_config_file())}",
        f"providers\t{','.join(sorted(SURFACES))}",
        f"next_step\trun `{slack_provider.config_command()}`",
    ]
    return "\n".join(lines)


def _render_surface_text(payload: dict[str, Any]) -> str:
    commands = payload.get("commands") or {}
    lines = [
        f"provider\t{payload.get('provider') or ''}",
        f"workspace\t{payload.get('workspace') or ''}",
        f"selected_workspace\t{payload.get('selectedWorkspace') or ''}",
        f"config_file\t{payload.get('configPathDisplay') or display_path(primary_config_file())}",
        f"persisted\t{str(bool(payload.get('persisted'))).lower()}",
        f"ready\t{str(bool(payload.get('ready'))).lower()}",
        f"slackdump_present\t{str(bool(payload.get('slackdumpPresent'))).lower()}",
        f"workspace_auth_configured\t{str(bool(payload.get('slackdumpAuthConfigured'))).lower()}",
        f"live_search_auth_configured\t{str(bool(payload.get('liveSearchAuthConfigured'))).lower()}",
        f"live_search_auth_usable\t{str(bool(payload.get('liveSearchAuthUsable'))).lower()}",
    ]
    derived_from = str(payload.get("derivedFrom") or "").strip()
    if derived_from:
        lines.append(f"derived_from\t{derived_from}")
    setup_error = str(payload.get("setupError") or "").strip()
    if setup_error:
        lines.append(f"setup_error\t{setup_error}")
    lines.append(f"next_step\t{payload.get('nextStep') or ''}")
    for key in ("config", "auth", "status", "workspaces"):
        value = str(commands.get(key) or "").strip()
        if value:
            lines.append(f"{key}_command\t{value}")
    return "\n".join(lines)


def _configure_surface(
    surface: ConfigSurface,
    *,
    target: str,
) -> tuple[dict[str, Any], int]:
    raw_target = target.strip()
    persisted = False
    auth_attempted = False
    setup_error = ""
    if raw_target:
        workspace = surface.resolve_target(raw_target)
        surface.persist(workspace)
        persisted = True
        if is_interactive():
            auth_attempted = True
            try:
                surface.auto_configure(workspace)
            except slack_provider.SlackError as exc:
                setup_error = str(exc)
    else:
        workspace = ""
    payload = dict(surface.inspect(workspace))
    payload.update(
        {
            "provider": surface.name,
            "configFile": str(primary_config_file()),
            "configPathDisplay": display_path(primary_config_file()),
            "persisted": persisted,
            "derivedFrom": raw_target,
            "authAttempted": auth_attempted,
            "setupError": setup_error,
        }
    )
    exit_code = 1 if setup_error else 0
    return payload, exit_code


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="gotta config",
        description=(
            "Persist durable provider defaults and guide exact readiness/auth setup. "
            "Prototype provider: Slack."
        ),
    )
    parser.add_argument("provider", nargs="?", choices=sorted(SURFACES))
    parser.add_argument(
        "target",
        nargs="?",
        help="provider-specific durable target, such as a workspace slug or a provider permalink",
    )
    parser.add_argument("--output", choices=["json", "text"], default="text")
    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    if is_long_help_request(argv):
        return print_long_help(parser)
    try:
        args = parser.parse_args(argv)
        if not args.provider:
            payload = {
                "configFile": str(primary_config_file()),
                "configPathDisplay": display_path(primary_config_file()),
                "providers": sorted(SURFACES),
                "nextStep": f"run `{slack_provider.config_command()}`",
            }
            if args.output == "json":
                print_json(payload)
            else:
                print(_render_summary_text(payload))
            return 0
        payload, exit_code = _configure_surface(
            SURFACES[args.provider], target=str(args.target or "")
        )
        if args.output == "json":
            print_json(payload)
        else:
            print(_render_surface_text(payload))
        return exit_code
    except ToolError as exc:
        return die(str(exc), code=1)
    except slack_provider.SlackError as exc:
        return die(str(exc), code=1)



"""Actor bind and bind-receipt helpers."""

from __future__ import annotations

from pathlib import Path

from gotta.content.path import sh_quote
from gotta.session.activity import _actor_log_line, _append_actor_event
from gotta.session.bootstrap import _ensure_actor_surface
from gotta.session.registry import (
    WANT_FILE,
    _actor_charter_command,
    _actor_dir_path,
    _actor_is_selected,
    _actor_label,
    _bind_actor_identity,
    _read_actor_state,
    _write_actor_state,
)
from gotta.session.status.blocker import _actor_launch_blockers


def _actor_launch_command(work_dir: Path, actor_name: str) -> str:
    return f"gotta actor launch {actor_name} --session {sh_quote(str(work_dir))}"


def _render_actor_bind_message(payload: dict[str, object]) -> str:
    actor = str(payload.get("actor") or "").strip()
    label = str(payload.get("label") or actor).strip()
    actor_want = str(payload.get("wantPath") or "").strip()
    actor_goal = str(payload.get("goalPath") or "").strip()
    want_cmd = str(payload.get("wantCommand") or "").strip()
    goal_cmd = str(payload.get("goalCommand") or "").strip()
    todo_cmd = str(payload.get("todoCommand") or "").strip()
    launch_cmd = str(payload.get("launchCommand") or "").strip()
    raw_blockers = payload.get("blockers")
    actor_blockers = (
        [str(item).strip() for item in raw_blockers if str(item).strip()]
        if isinstance(raw_blockers, list)
        else []
    )
    if actor_blockers:
        suffix = (
            f"; not launched. This bind completed the addressable actor target, so rewrite `{actor_want}` and `{actor_goal}` for {label} with `{want_cmd}` and `{goal_cmd}` first. "
            f"Use `{todo_cmd}` to extend the minimal actor-local checklist before launch if useful, "
            f"then launch with `{launch_cmd}` when you actually want {label} to start"
        )
    else:
        suffix = (
            f"; not launched. `{actor_want}` and `{actor_goal}` are already real. "
            f"Use `{todo_cmd}` to extend the minimal actor-local checklist before launch if useful, "
            f"then launch with `{launch_cmd}` when you actually want {label} to start"
        )
    if bool(payload.get("alreadyBound")):
        return f"{actor} ({label}) already bound{suffix}"
    created_note = " [new actor]" if bool(payload.get("created")) else ""
    return (
        f"bound {actor} ({label}) session, seeded actor-local WANT/GOAL placeholders, minimal actor-local canonical state, and shared evidence access{created_note}"
        f"{suffix}"
    )


def _bind_actor(session_root: Path, actor_name: str) -> dict[str, object]:
    actor, created = _bind_actor_identity(session_root, actor_name)
    already_bound = _actor_is_selected(session_root, actor)
    _ensure_actor_surface(session_root, actor)
    label = _actor_label(actor, work_dir=session_root)
    current_status = str(
        _read_actor_state(session_root, actor).get("status") or "pending"
    )
    if current_status in {
        "",
        "pending",
        "completed",
        "failed",
        "incomplete",
        "rejected",
        "signed_off",
    }:
        _write_actor_state(session_root, actor, {"status": "bound"})
    launch_cmd = _actor_launch_command(session_root, actor)
    actor_want = _actor_dir_path(session_root, actor) / WANT_FILE
    actor_goal = _actor_dir_path(session_root, actor) / "GOAL.md"
    want_cmd = _actor_charter_command(actor, "want")
    goal_cmd = _actor_charter_command(actor, "goal")
    actor_blockers = _actor_launch_blockers(session_root, actor_name=actor)
    todo_cmd = f"gotta todo --actor {actor}"
    if already_bound:
        status = "already_bound"
    else:
        _append_actor_event(
            session_root, actor, event="bound", detail="bound actor session"
        )
        _actor_log_line(session_root, actor, "bound session")
        status = "bound"
    payload: dict[str, object] = {
        "actor": actor,
        "label": label,
        "created": created,
        "alreadyBound": already_bound,
        "status": status,
        "sessionRoot": str(session_root.resolve()),
        "actorRoot": str(_actor_dir_path(session_root, actor).resolve()),
        "wantPath": str(actor_want.resolve()),
        "goalPath": str(actor_goal.resolve()),
        "wantCommand": want_cmd,
        "goalCommand": goal_cmd,
        "todoCommand": todo_cmd,
        "launchCommand": launch_cmd,
        "blockers": actor_blockers,
    }
    payload["message"] = _render_actor_bind_message(payload)
    return payload
